import abjad
import math


def make_desordre_pitches():
    """
    Makes Désordre pitches.
    """

    right_hand = [
        [[-1, 4, 5], [-1, 4, 5, 7, 9]],
        [[0, 7, 9], [-1, 4, 5, 7, 9]],
        [[2, 4, 5, 7, 9], [0, 5, 7]],
        [[-3, -1, 0, 2, 4, 5, 7]],
        [[-3, 2, 4], [-3, 2, 4, 5, 7]],
        [[2, 5, 7], [-3, 9, 11, 12, 14]],
        [[4, 5, 7, 9, 11], [2, 4, 5]],
        [[-5, 4, 5, 7, 9, 11, 12]],
        [[2, 9, 11], [2, 9, 11, 12, 14]],
        ]

    left_hand = [
        [[-9, -4, -2], [-9, -4, -2, 1, 3]],
        [[-6, -2, 1], [-9, -4, -2, 1, 3]],
        [[-4, -2, 1, 3, 6], [-4, -2, 1]],
        [[-9, -6, -4, -2, 1, 3, 6, 1]],
        [[-6, -2, 1], [-6, -2, 1, 3, -2]],
        [[-4, 1, 3], [-6, 3, 6, -6, -4]],
        [[-14, -11, -9, -6, -4], [-14, -11, -9]],
        [[-11, -2, 1, -6, -4, -2, 1, 3]],
        [[-6, 1, 3], [-6, -4, -2, 1, 3]],
        ]

    return [right_hand, left_hand]


def make_desordre_cell(pitches):
    """
    Makes a Désordre cell.
    """

    notes = [abjad.Note(pitch, (1, 8)) for pitch in pitches]
    notes = abjad.Selection(notes)
    beam = abjad.Beam()
    abjad.attach(beam, notes)
    slur = abjad.Slur()
    abjad.attach(slur, notes)
    clef = abjad.Dynamic('f')
    abjad.attach(clef, notes[0])
    dynamic = abjad.Dynamic('p')
    abjad.attach(dynamic, notes[1])

    # make the lower voice
    lower_voice = abjad.Voice(notes)
    lower_voice.name = 'RH Lower Voice'
    command = abjad.LilyPondLiteral(r'\voiceTwo')
    abjad.attach(command, lower_voice)
    n = int(math.ceil(len(pitches) / 2.))
    chord = abjad.Chord([pitches[0], pitches[0] + 12], (n, 8))
    articulation = abjad.Articulation('>')
    abjad.attach(articulation, chord)

    # make the upper voice
    upper_voice = abjad.Voice([chord])
    upper_voice.name = 'RH Upper Voice'
    command = abjad.LilyPondLiteral(r'\voiceOne')
    abjad.attach(command, upper_voice)

    # combine them together
    container = abjad.Container([lower_voice, upper_voice])
    container.is_simultaneous = True

    # make all 1/8 beats breakable
    leaves = abjad.select(lower_voice).leaves()
    for leaf in leaves[:-1]:
        bar_line = abjad.BarLine('')
        abjad.attach(bar_line, leaf)

    return container


def make_desordre_measure(pitches):
    """
    Makes a measure composed of *Désordre cells*.

    ``pitches`` is a list of lists of number (e.g., [[1, 2, 3], [2, 3, 4]])

    Returns a measure.
    """
    for sequence in pitches:
        container = make_desordre_cell(sequence)
        time_signature = abjad.inspect(container).duration()
        time_signature = abjad.NonreducedFraction(time_signature)
        time_signature = time_signature.with_denominator(8)
        measure = abjad.Measure(time_signature, [container])
    return measure


def make_desordre_staff(pitches):
    """
    Makes Désordre staff.
    """

    staff = abjad.Staff()
    for sequence in pitches:
        measure = make_desordre_measure(sequence)
        staff.append(measure)
    return staff


def make_desordre_score(pitches):
    """
    Makes Désordre score.
    """

    assert len(pitches) == 2
    staff_group = abjad.StaffGroup()
    staff_group.lilypond_type = 'PianoStaff'

    # build the music
    for hand in pitches:
        staff = make_desordre_staff(hand)
        staff_group.append(staff)

    # set clef and key signature to left hand staff
    leaf = abjad.inspect(staff_group[1]).leaf(0)
    clef = abjad.Clef('bass')
    abjad.attach(clef, leaf)
    key_signature = abjad.KeySignature('b', 'major')
    abjad.attach(key_signature, leaf)

    # wrap the piano staff in a score
    score = abjad.Score([staff_group])

    return score


def make_desordre_lilypond_file(score):
    """
    Makes Désordre LilyPond file.
    """
    lilypond_file = abjad.LilyPondFile.new(
        music=score,
        default_paper_size=('a4', 'letter'),
        global_staff_size=14,
        )

    lilypond_file.layout_block.indent = 0
    lilypond_file.layout_block.ragged_right = True
    lilypond_file.layout_block.merge_differently_dotted = True
    lilypond_file.layout_block.merge_differently_headed = True

    context_block = abjad.ContextBlock(
        source_lilypond_type='Score',
        )
    lilypond_file.layout_block.items.append(context_block)
    context_block.remove_commands.append('Bar_number_engraver')
    context_block.remove_commands.append('Default_bar_line_engraver')
    context_block.remove_commands.append('Timing_translator')
    abjad.override(context_block).beam.breakable = True
    abjad.override(context_block).glissando.breakable = True
    abjad.override(context_block).note_column.ignore_collision = True
    abjad.override(context_block).spacing_spanner.uniform_stretching = True
    abjad.override(context_block).text_script.staff_padding = 4
    abjad.override(context_block).text_spanner.breakable = True
    abjad.override(context_block).tuplet_bracket.bracket_visibility = True
    abjad.override(context_block).tuplet_bracket.minimum_length = 3
    abjad.override(context_block).tuplet_bracket.padding = 2
    scheme = abjad.Scheme('ly:spanner::set-spacing-rods')
    abjad.override(context_block).tuplet_bracket.springs_and_rods = scheme
    scheme = abjad.Scheme('tuplet-number::calc-fraction-text')
    abjad.override(context_block).tuplet_number.text = scheme
    abjad.setting(context_block).autoBeaming = False
    moment = abjad.SchemeMoment((1, 12))
    abjad.setting(context_block).proportionalNotationDuration = moment
    abjad.setting(context_block).tupletFullLength = True

    context_block = abjad.ContextBlock(
        source_lilypond_type='Staff',
        )
    lilypond_file.layout_block.items.append(context_block)
    # LilyPond CAUTION: Timing_translator must appear
    #                   before Default_bar_line_engraver!
    context_block.consists_commands.append('Timing_translator')
    context_block.consists_commands.append('Default_bar_line_engraver')
    scheme = abjad.Scheme("'numbered")
    abjad.override(context_block).time_signature.style = scheme

    context_block = abjad.ContextBlock(
        source_lilypond_type='RhythmicStaff',
        )
    lilypond_file.layout_block.items.append(context_block)
    # LilyPond CAUTION: Timing_translator must appear
    #                   before Default_bar_line_engraver!
    context_block.consists_commands.append('Timing_translator')
    context_block.consists_commands.append('Default_bar_line_engraver')
    scheme = abjad.Scheme("'numbered")
    abjad.override(context_block).time_signature.style = scheme
    abjad.override(context_block).vertical_axis_group.minimum_Y_extent = (-2, 4)

    context_block = abjad.ContextBlock(
        source_lilypond_type='Voice',
        )
    lilypond_file.layout_block.items.append(context_block)
    context_block.remove_commands.append('Forbid_line_break_engraver')

    return lilypond_file


if __name__ == '__main__':
    from abjad import show
    pitches = make_desordre_pitches()
    score = make_desordre_score(pitches)
    lilypond_file = make_desordre_lilypond_file(score)
    show(lilypond_file)



from abjad import Fraction


def fraction_to_proper_fraction(rational):
    """
    Changes ``rational`` to proper fraction.

    ..  container:: example

        >>> abjad.mathtools.fraction_to_proper_fraction(abjad.Fraction(116, 8))
        (14, Fraction(1, 2))

    Returns pair.
    """
    assert isinstance(rational, Fraction), repr(rational)
    quotient = int(rational)
    residue = rational - quotient
    return quotient, residue



from abjad.system.AbjadValueObject import AbjadValueObject


class Infinity(AbjadValueObject):
    """
    Infinity.

    ..  container:: example

        All numbers compare less than infinity:

        >>> 9999999 < Infinity
        True

        >>> 2**38 < Infinity
        True

    ..  container:: example

        Infinity compares equal to itself:

        >>> Infinity == Infinity
        True

    ..  container:: example

        Negative infinity compares less than infinity:

        >>> NegativeInfinity < Infinity
        True

    Initializes as a system singleton at start-up.

    Available as a built-in after Abjad starts.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_value',
        )

    ### INTIALIZER ###

    def __init__(self):
        self._value = float('infinity')

    ### SPECIAL METHODS ###

    def __eq__(self, argument):
        """
        Is true when ``argument`` is also infinity.

        Returns true or false.
        """
        return super().__eq__(argument)

    def __float__(self):
        """
        Convert infinity to float.

        Returns float.
        """
        return self._value

    def __ge__(self, argument):
        """
        Is true for all values of ``argument``.

        Returns true.
        """
        return self._value >= argument

    def __gt__(self, argument):
        """
        Is true for all noninfinite values of ``argument``.

        Returns true or false.
        """
        return self._value > argument

    def __hash__(self):
        """
        Hashes infinity.

        Required to be explicitly redefined on Python 3 if __eq__ changes.

        Returns integer.
        """
        return super().__hash__()

    def __le__(self, argument):
        """
        Is true when ``argument`` is infinite.

        Returns true or false.
        """
        return self._value <= argument

    def __lt__(self, argument):
        """
        Is true for no values of ``argument``.

        Returns true or false.
        """
        return self._value < argument

    def __sub__(self, argument):
        """
        Subtracts ``argument`` from infinity.

        Returns infinity or 0 if ``argument`` is also infinity.
        """
        if argument is self:
            return 0
        return self

    ### PRIVATE METHODS ###

    def _get_format_specification(self):
        from abjad import system
        return system.FormatSpecification(
            client=self,
            repr_text=type(self).__name__,
            storage_format_text=type(self).__name__,
            )



from abjad import Left, Right
from abjad import utilities
from abjad import mathtools
from abjad import pitch as abjad_pitch
from abjad import core
from abjad import spanners
from abjad.top import attach
from abjad.top import detach
from abjad.system import Parser


class ReducedLyParser(Parser):
    r"""
    Parses the "reduced-ly" syntax, a modified subset of LilyPond syntax.

    ..  container:: example

        >>> parser = abjad.parser.ReducedLyParser()

        Understands LilyPond-like representation of notes, chords and rests:

        >>> string = "c'4 r8. <b d' fs'>16"
        >>> container = parser(string)
        >>> abjad.show(container) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(container)
            {
                c'4
                r8.
                <b d' fs'>16
            }

        Also parses bare duration as notes on middle-C, and negative bare
        durations as rests:

        >>> string = '4 -8 16. -32'
        >>> container = parser(string)
        >>> abjad.show(container) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(container)
            {
                c'4
                r8
                c'16.
                r32
            }

    ..  container:: example

        Note that the leaf syntax is greedy, and therefore duration specifiers
        following pitch specifiers will be treated as part of the same
        expression. The following produces 2 leaves, rather than 3:

        >>> string = "4 d' 4"
        >>> container = parser(string)
        >>> abjad.show(container) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(container)
            {
                c'4
                d'4
            }

        Understands LilyPond-like default durations:

        >>> string = "c'4 d' e' f'"
        >>> container = parser(string)
        >>> abjad.show(container) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(container)
            {
                c'4
                d'4
                e'4
                f'4
            }

    Also understands various types of container specifications.

    ..  container:: example

        Can create arbitrarily nested tuplets:

        >>> string = "2/3 { 4 4 3/5 { 8 8 8 } }"
        >>> tuplet = parser(string)
        >>> abjad.show(tuplet) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(tuplet)
            \tweak edge-height #'(0.7 . 0)
            \times 2/3 {
                c'4
                c'4
                \tweak text #tuplet-number::calc-fraction-text
                \tweak edge-height #'(0.7 . 0)
                \times 3/5 {
                    c'8
                    c'8
                    c'8
                }
            }

    ..  container:: example

        Can create measures too:

        >>> string = '| 4/4 4 4 4 4 || 3/8 8 8 8 |'
        >>> container = parser(string)
        >>> abjad.show(container) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(container)
            {
                {   % measure
                    \time 4/4
                    c'4
                    c'4
                    c'4
                    c'4
                }   % measure
                {   % measure
                    \time 3/8
                    c'8
                    c'8
                    c'8
                }   % measure
            }

    ..  container:: example

        Finally, understands ties, slurs and beams:

        >>> string = 'c16 [ ( d ~ d ) f ]'
        >>> container = parser(string)
        >>> abjad.show(container) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(container)
            {
                c16
                [
                (
                d16
                ~
                d16
                )
                f16
                ]
            }

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_default_duration',
        '_toplevel_component_count',
        )

    ### INITIALIZER ###

    def __init__(self, debug=False):
        self._default_duration = utilities.Duration((1, 4))
        self._toplevel_component_count = None
        Parser.__init__(self, debug=debug)

    ### LEX SETUP ###

    tokens = (
        'APOSTROPHE',
        'BRACE_L',
        'BRACE_R',
        'BRACKET_L',
        'BRACKET_R',
        'CARAT_L',
        'CARAT_R',
        'COMMA',
        'DOT',
        'FRACTION',
        'INTEGER_N',
        'INTEGER_P',
        'PAREN_L',
        'PAREN_R',
        'PIPE',
        'PITCHNAME',
        'RESTNAME',
        'TILDE',
    )

    t_APOSTROPHE = "'"
    t_BRACE_L = '{'
    t_BRACE_R = '}'
    t_BRACKET_L = '\['
    t_BRACKET_R = '\]'
    t_CARAT_L = '\<'
    t_CARAT_R = '\>'
    t_COMMA = ','
    t_DOT = '\.'
    t_PAREN_L = '\('
    t_PAREN_R = '\)'
    t_PIPE = '\|'
    t_RESTNAME = 'r'
    t_TILDE = '~'

    t_ignore = ' \t\r'

    ### YACC SETUP ###

    start = 'start'

    ### LEX METHODS ###

    def t_FRACTION(self, t):
        r'([1-9]\d*/[1-9]\d*)'
        parts = t.value.split('/')
        t.value = mathtools.NonreducedFraction(int(parts[0]), int(parts[1]))
        return t

    def t_INTEGER_N(self, t):
        r'(-[1-9]\d*)'
        t.value = int(t.value)
        return t

    def t_INTEGER_P(self, t):
        r'([1-9]\d*)'
        t.value = int(t.value)
        return t

    def t_PITCHNAME(self, t):
        r'[a-g](ff|ss|f|s|tqf|tqs|qs|qf)?'
        t.value = abjad_pitch.NamedPitchClass(t.value)
        return t

    def t_error(self, t):
        print(("Illegal character '%s'" % t.value[0]))
        t.lexer.skip(1)

    def t_newline(self, t):
        r'\n+'
        t.lexer.lineno += t.value.count('\n')

    ### YACC METHODS ###

    def p_apostrophes__APOSTROPHE(self, p):
        """
        apostrophes : APOSTROPHE
        """
        p[0] = 1

    def p_apostrophes__apostrophes__APOSTROPHE(self, p):
        """
        apostrophes : apostrophes APOSTROPHE
        """
        p[0] = p[1] + 1

    def p_beam__BRACKET_L(self, p):
        """
        beam : BRACKET_L
        """
        import abjad
        p[0] = (abjad.Beam, Left)

    def p_beam__BRACKET_R(self, p):
        """
        beam : BRACKET_R
        """
        import abjad
        p[0] = (abjad.Beam, Right)

    def p_chord_body__chord_pitches(self, p):
        """
        chord_body : chord_pitches
        """
        p[0] = core.Chord(p[1], self._default_duration)

    def p_chord_body__chord_pitches__positive_leaf_duration(self, p):
        """
        chord_body : chord_pitches positive_leaf_duration
        """
        p[0] = core.Chord(p[1], p[2])

    def p_chord_pitches__CARAT_L__pitches__CARAT_R(self, p):
        """
        chord_pitches : CARAT_L pitches CARAT_R
        """
        p[0] = p[2]

    def p_commas__COMMA(self, p):
        """
        commas : COMMA
        """
        p[0] = 1

    def p_commas__commas__commas(self, p):
        """
        commas : commas COMMA
        """
        p[0] = p[1] + 1

    def p_component__container(self, p):
        """
        component : container
        """
        p[0] = p[1]

    def p_component__fixed_duration_container(self, p):
        """
        component : fixed_duration_container
        """
        p[0] = p[1]

    def p_component__leaf(self, p):
        """
        component : leaf
        """
        p[0] = p[1]

    def p_component__tuplet(self, p):
        """
        component : tuplet
        """
        p[0] = p[1]

    def p_component_list__EMPTY(self, p):
        """
        component_list :
        """
        p[0] = []

    def p_component_list__component_list__component(self, p):
        """
        component_list : component_list component
        """
        p[0] = p[1] + [p[2]]

    def p_container__BRACE_L__component_list__BRACE_R(self, p):
        r"""container : BRACE_L component_list BRACE_R
        """
        p[0] = core.Container()
        for component in p[2]:
            p[0].append(component)

    def p_dots__EMPTY(self, p):
        """
        dots :
        """
        p[0] = 0

    def p_dots__dots__DOT(self, p):
        """
        dots : dots DOT
        """
        p[0] = p[1] + 1

    def p_error(self, p):
        if p:
            print(("Syntax error at '%s'" % p.value))
        else:
            print("Syntax error at EOF")

    def p_fixed_duration_container__BRACE_L__FRACTION__BRACE_R(self, p):
        """
        fixed_duration_container : BRACE_L FRACTION BRACE_R
        """
        raise Exception('fixed-duration containers no longer supported.')

    def p_leaf__leaf_body__post_events(self, p):
        """
        leaf : leaf_body post_events
        """
        p[0] = p[1]
        if p[2]:
            annotation = {'post events': p[2]}
            attach(annotation, p[0])

    def p_leaf_body__chord_body(self, p):
        """
        leaf_body : chord_body
        """
        p[0] = p[1]

    def p_leaf_body__note_body(self, p):
        """
        leaf_body : note_body
        """
        p[0] = p[1]

    def p_leaf_body__rest_body(self, p):
        """
        leaf_body : rest_body
        """
        p[0] = p[1]

    def p_measure__PIPE__FRACTION__component_list__PIPE(self, p):
        """
        measure : PIPE FRACTION component_list PIPE
        """
        measure = core.Measure(p[2].pair)
        for x in p[3]:
            measure.append(x)
        p[0] = measure

    def p_negative_leaf_duration__INTEGER_N__dots(self, p):
        """
        negative_leaf_duration : INTEGER_N dots
        """
        duration_log = p[1]
        dots = '.' * p[2]
        duration = utilities.Duration.from_lilypond_duration_string(
            '{}{}'.format(abs(duration_log), dots))
        self._default_duration = duration
        p[0] = duration

    def p_note_body__pitch(self, p):
        """
        note_body : pitch
        """
        p[0] = core.Note(p[1], self._default_duration)

    def p_note_body__pitch__positive_leaf_duration(self, p):
        """
        note_body : pitch positive_leaf_duration
        """
        p[0] = core.Note(p[1], p[2])

    def p_note_body__positive_leaf_duration(self, p):
        """
        note_body : positive_leaf_duration
        """
        p[0] = core.Note(0, p[1])

    def p_pitch__PITCHNAME(self, p):
        """
        pitch : PITCHNAME
        """
        p[0] = abjad_pitch.NamedPitch(str(p[1]))

    def p_pitch__PITCHNAME__apostrophes(self, p):
        """
        pitch : PITCHNAME apostrophes
        """
        p[0] = abjad_pitch.NamedPitch(str(p[1]) + "'" * p[2])

    def p_pitch__PITCHNAME__commas(self, p):
        """
        pitch : PITCHNAME commas
        """
        p[0] = abjad_pitch.NamedPitch(str(p[1]) + ',' * p[2])

    def p_pitches__pitch(self, p):
        """
        pitches : pitch
        """
        p[0] = [p[1]]

    def p_pitches__pitches__pitch(self, p):
        """
        pitches : pitches pitch
        """
        p[0] = p[1] + [p[2]]

    def p_positive_leaf_duration__INTEGER_P__dots(self, p):
        """
        positive_leaf_duration : INTEGER_P dots
        """
        duration_log = p[1]
        dots = '.' * p[2]
        duration = utilities.Duration.from_lilypond_duration_string(
            '{}{}'.format(abs(duration_log), dots))
        self._default_duration = duration
        p[0] = duration

    def p_post_event__beam(self, p):
        """
        post_event : beam
        """
        p[0] = p[1]

    def p_post_event__slur(self, p):
        """
        post_event : slur
        """
        p[0] = p[1]

    def p_post_event__tie(self, p):
        """
        post_event : tie
        """
        p[0] = p[1]

    def p_post_events__EMPTY(self, p):
        """
        post_events :
        """
        p[0] = {}

    def p_post_events__post_events__post_event(self, p):
        """
        post_events : post_events post_event
        """
        kind, direction = p[2]
        if kind in p[1]:
            p[1][kind].append(direction)
        else:
            p[1][kind] = [direction]
        p[0] = p[1]

    def p_rest_body__RESTNAME(self, p):
        """
        rest_body : RESTNAME
        """
        p[0] = core.Rest(self._default_duration)

    def p_rest_body__RESTNAME__positive_leaf_duration(self, p):
        """
        rest_body : RESTNAME positive_leaf_duration
        """
        p[0] = core.Rest(p[2])

    def p_rest_body__negative_leaf_duration(self, p):
        """
        rest_body : negative_leaf_duration
        """
        p[0] = core.Rest(p[1])

    def p_slur__PAREN_L(self, p):
        """
        slur : PAREN_L
        """
        import abjad
        p[0] = (abjad.Slur, Left)

    def p_slur__PAREN_R(self, p):
        """
        slur : PAREN_R
        """
        import abjad
        p[0] = (abjad.Slur, Right)

    def p_start__EMPTY(self, p):
        """
        start :
        """
        self._toplevel_component_count = 0
        p[0] = []

    def p_start__start__component(self, p):
        """
        start : start component
        """
        self._toplevel_component_count += 1
        p[0] = p[1] + [p[2]]

    def p_start__start__measure(self, p):
        """
        start : start measure
        """
        self._toplevel_component_count += 1
        p[0] = p[1] + [p[2]]

    def p_tie__TILDE(self, p):
        """
        tie : TILDE
        """
        import abjad
        p[0] = (abjad.Tie, Left)

    def p_tuplet__FRACTION__container(self, p):
        """
        tuplet : FRACTION container
        """
        assert isinstance(p[2], core.Container)
        leaves = p[2][:]
        p[2][:] = []
        p[0] = core.Tuplet(p[1], leaves)

    ### PRIVATE METHODS ###

    def _apply_spanners(self, leaves):
        import abjad

        spanner_references = {
            abjad.Beam: None,
            abjad.Slur: None,
        }

        first_leaf = leaves[0]
        pairs = abjad.sequence(leaves).nwise(wrapped=True)
        for leaf, next_leaf in pairs:
            span_events = self._get_span_events(leaf)
            for current_class, directions in span_events.items():
                starting, stopping = [], []
                for direction in directions:
                    if direction is Left:
                        starting.append(Left)
                    else:
                        stopping.append(Right)

                # apply undirected events immediately,
                # and do not maintain a reference to them
                if current_class is abjad.Tie:
                    if next_leaf is first_leaf:
                        message = 'unterminated {} at {}.'
                        message = message.format(current_class.__name__, leaf)
                        raise Exception(message)
                    previous_tie = [
                        x for x in leaf._get_spanners()
                        if isinstance(x, abjad.Tie)
                        ]
                    if previous_tie:
                        previous_tie[0]._append(next_leaf)
                    else:
                        tie = abjad.Tie()
                        selection = abjad.select([leaf, next_leaf])
                        attach(tie, selection)

                elif current_class is abjad.Beam:
                    # A beam may begin and end on the same leaf
                    # but only one beam spanner may cover any given leaf,
                    # and starting events are processed before ending ones
                    for _ in starting:
                        if spanner_references[current_class] is not None:
                            message = 'already have beam.'
                            raise Exception(message)
                        else:
                            spanner_references[current_class] = current_class()
                    for _ in stopping:
                        if spanner_references[current_class] is not None:
                            spanner_references[current_class]._append(leaf)
                            spanner_references[current_class] = None

                elif current_class is spanners.Slur:
                    # Slurs process stop events before start events,
                    # they must contain more than one leaf,
                    # but they can stop on a leaf and start on the same leaf.
                    for _ in stopping:
                        if spanner_references[current_class] is not None:
                            spanner_references[current_class]._append(leaf)
                            spanner_references[current_class] = None
                        else:
                            message = 'can not end: {}.'
                            message = message.format(current_class.__name)
                            raise Exception(message)
                    for _ in starting:
                        if spanner_references[current_class] is None:
                            spanner_references[current_class] = current_class()
                        else:
                            message = 'already have: {}.'
                            message = message.format(current_class.__name)
                            raise Exception(message)

            # append leaf to all tracked spanners,
            for current_class, instance in spanner_references.items():
                if instance is not None:
                    instance._append(leaf)

        # check for unterminated spanners
        for current_class, instance in spanner_references.items():
            if instance is not None:
                message = 'unterminated {}.'
                message = message.format(current_class.__name__)
                raise Exception(message)

    def _cleanup(self, parsed):
        import abjad
        container = core.Container()
        for x in parsed:
            container.append(x)
        parsed = container
        leaves = abjad.select(parsed).leaves()
        if leaves:
            self._apply_spanners(leaves)
        for leaf in leaves:
            detach(dict, leaf)
        if 1 < self._toplevel_component_count:
            return parsed
        return parsed[0]

    def _get_span_events(self, leaf):
        annotations = leaf._get_indicators(dict)
        detach(dict, leaf)
        annotations = [x for x in annotations if 'post events' in x]
        if annotations:
            return annotations[0]['post events']
        return {}

    def _setup(self):
        self._toplevel_component_count = 0
        self._default_duration = utilities.Duration((1, 4))

    ### PUBLIC PROPERTIES ###

    @property
    def debug(self):
        """
        Gets debug boolean of reduced ly parser.

        Returns true or false.
        """
        return self._debug

    @property
    def lexer_rules_object(self):
        """
        Lexer rules object of reduced ly parser.
        """
        return self

    @property
    def parser_rules_object(self):
        """
        Parser rules object of reduced ly parser.
        """
        return self



import collections
import typing
from abjad import enums
from abjad.markups import Markup
from abjad.mathtools.Ratio import Ratio
from abjad.system.AbjadValueObject import AbjadValueObject
from abjad.system.LilyPondFormatBundle import LilyPondFormatBundle
from abjad.top.inspect import inspect
from abjad.top.new import new
from abjad.top.select import select
from abjad.utilities.Duration import Duration


class MetricModulation(AbjadValueObject):
    r"""
    Metric modulation.

    ..  container:: example

        With notes:

        >>> metric_modulation = abjad.MetricModulation(
        ...     left_rhythm=abjad.Note("c'4"),
        ...     right_rhythm=abjad.Note("c'4."),
        ...     )

        >>> abjad.show(metric_modulation) # doctest: +SKIP

        ..  docs::

            >>> print(format(metric_modulation, 'lilypond'))
            \markup {
                \score
                    {
                        \new Score
                        \with
                        {
                            \override SpacingSpanner.spacing-increment = #0.5
                            proportionalNotationDuration = ##f
                        }
                        <<
                            \new RhythmicStaff
                            \with
                            {
                                \remove Time_signature_engraver
                                \remove Staff_symbol_engraver
                                \override Stem.direction = #up
                                \override Stem.length = #5
                                \override TupletBracket.bracket-visibility = ##t
                                \override TupletBracket.direction = #up
                                \override TupletBracket.minimum-length = #4
                                \override TupletBracket.padding = #1.25
                                \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                \override TupletNumber.font-size = #0
                                \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                tupletFullLength = ##t
                            }
                            {
                                c'4
                            }
                        >>
                        \layout {
                            indent = #0
                            ragged-right = ##t
                        }
                    }
                =
                \hspace
                    #-0.5
                \score
                    {
                        \new Score
                        \with
                        {
                            \override SpacingSpanner.spacing-increment = #0.5
                            proportionalNotationDuration = ##f
                        }
                        <<
                            \new RhythmicStaff
                            \with
                            {
                                \remove Time_signature_engraver
                                \remove Staff_symbol_engraver
                                \override Stem.direction = #up
                                \override Stem.length = #5
                                \override TupletBracket.bracket-visibility = ##t
                                \override TupletBracket.direction = #up
                                \override TupletBracket.minimum-length = #4
                                \override TupletBracket.padding = #1.25
                                \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                \override TupletNumber.font-size = #0
                                \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                tupletFullLength = ##t
                            }
                            {
                                c'4.
                            }
                        >>
                        \layout {
                            indent = #0
                            ragged-right = ##t
                        }
                    }
                }

    ..  container:: example

        With tuplets:

        >>> metric_modulation = abjad.MetricModulation(
        ...     left_rhythm=abjad.Tuplet((4, 5), "c'4"),
        ...     right_rhythm=abjad.Note("c'4"),
        ...     )

        >>> abjad.show(metric_modulation) # doctest: +SKIP

        ..  docs::

            >>> print(format(metric_modulation, 'lilypond'))
            \markup {
                \score
                    {
                        \new Score
                        \with
                        {
                            \override SpacingSpanner.spacing-increment = #0.5
                            proportionalNotationDuration = ##f
                        }
                        <<
                            \new RhythmicStaff
                            \with
                            {
                                \remove Time_signature_engraver
                                \remove Staff_symbol_engraver
                                \override Stem.direction = #up
                                \override Stem.length = #5
                                \override TupletBracket.bracket-visibility = ##t
                                \override TupletBracket.direction = #up
                                \override TupletBracket.minimum-length = #4
                                \override TupletBracket.padding = #1.25
                                \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                \override TupletNumber.font-size = #0
                                \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                tupletFullLength = ##t
                            }
                            {
                                \tweak edge-height #'(0.7 . 0)
                                \times 4/5 {
                                    c'4
                                }
                            }
                        >>
                        \layout {
                            indent = #0
                            ragged-right = ##t
                        }
                    }
                =
                \hspace
                    #-0.5
                \score
                    {
                        \new Score
                        \with
                        {
                            \override SpacingSpanner.spacing-increment = #0.5
                            proportionalNotationDuration = ##f
                        }
                        <<
                            \new RhythmicStaff
                            \with
                            {
                                \remove Time_signature_engraver
                                \remove Staff_symbol_engraver
                                \override Stem.direction = #up
                                \override Stem.length = #5
                                \override TupletBracket.bracket-visibility = ##t
                                \override TupletBracket.direction = #up
                                \override TupletBracket.minimum-length = #4
                                \override TupletBracket.padding = #1.25
                                \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                \override TupletNumber.font-size = #0
                                \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                tupletFullLength = ##t
                            }
                            {
                                c'4
                            }
                        >>
                        \layout {
                            indent = #0
                            ragged-right = ##t
                        }
                    }
                }

        >>> metric_modulation = abjad.MetricModulation(
        ...     left_rhythm=abjad.Note("c'4"),
        ...     right_rhythm=abjad.Tuplet((4, 5), "c'4"),
        ...     )

        >>> abjad.show(metric_modulation) # doctest: +SKIP

        ..  docs::

            >>> print(format(metric_modulation, 'lilypond'))
            \markup {
                \score
                    {
                        \new Score
                        \with
                        {
                            \override SpacingSpanner.spacing-increment = #0.5
                            proportionalNotationDuration = ##f
                        }
                        <<
                            \new RhythmicStaff
                            \with
                            {
                                \remove Time_signature_engraver
                                \remove Staff_symbol_engraver
                                \override Stem.direction = #up
                                \override Stem.length = #5
                                \override TupletBracket.bracket-visibility = ##t
                                \override TupletBracket.direction = #up
                                \override TupletBracket.minimum-length = #4
                                \override TupletBracket.padding = #1.25
                                \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                \override TupletNumber.font-size = #0
                                \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                tupletFullLength = ##t
                            }
                            {
                                c'4
                            }
                        >>
                        \layout {
                            indent = #0
                            ragged-right = ##t
                        }
                    }
                =
                \hspace
                    #-0.5
                \score
                    {
                        \new Score
                        \with
                        {
                            \override SpacingSpanner.spacing-increment = #0.5
                            proportionalNotationDuration = ##f
                        }
                        <<
                            \new RhythmicStaff
                            \with
                            {
                                \remove Time_signature_engraver
                                \remove Staff_symbol_engraver
                                \override Stem.direction = #up
                                \override Stem.length = #5
                                \override TupletBracket.bracket-visibility = ##t
                                \override TupletBracket.direction = #up
                                \override TupletBracket.minimum-length = #4
                                \override TupletBracket.padding = #1.25
                                \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                \override TupletNumber.font-size = #0
                                \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                tupletFullLength = ##t
                            }
                            {
                                \tweak edge-height #'(0.7 . 0)
                                \times 4/5 {
                                    c'4
                                }
                            }
                        >>
                        \layout {
                            indent = #0
                            ragged-right = ##t
                        }
                    }
                }

    ..  container:: example

        With tuplets again:

        >>> metric_modulation = abjad.MetricModulation(
        ...     left_rhythm=abjad.Note("c16."),
        ...     right_rhythm=abjad.Tuplet((2, 3), "c8"),
        ...     )

        >>> abjad.show(metric_modulation) # doctest: +SKIP

        ..  docs::

            >>> print(format(metric_modulation, 'lilypond'))
            \markup {
                \score
                    {
                        \new Score
                        \with
                        {
                            \override SpacingSpanner.spacing-increment = #0.5
                            proportionalNotationDuration = ##f
                        }
                        <<
                            \new RhythmicStaff
                            \with
                            {
                                \remove Time_signature_engraver
                                \remove Staff_symbol_engraver
                                \override Stem.direction = #up
                                \override Stem.length = #5
                                \override TupletBracket.bracket-visibility = ##t
                                \override TupletBracket.direction = #up
                                \override TupletBracket.minimum-length = #4
                                \override TupletBracket.padding = #1.25
                                \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                \override TupletNumber.font-size = #0
                                \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                tupletFullLength = ##t
                            }
                            {
                                c16.
                            }
                        >>
                        \layout {
                            indent = #0
                            ragged-right = ##t
                        }
                    }
                =
                \hspace
                    #-0.5
                \score
                    {
                        \new Score
                        \with
                        {
                            \override SpacingSpanner.spacing-increment = #0.5
                            proportionalNotationDuration = ##f
                        }
                        <<
                            \new RhythmicStaff
                            \with
                            {
                                \remove Time_signature_engraver
                                \remove Staff_symbol_engraver
                                \override Stem.direction = #up
                                \override Stem.length = #5
                                \override TupletBracket.bracket-visibility = ##t
                                \override TupletBracket.direction = #up
                                \override TupletBracket.minimum-length = #4
                                \override TupletBracket.padding = #1.25
                                \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                \override TupletNumber.font-size = #0
                                \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                tupletFullLength = ##t
                            }
                            {
                                \tweak edge-height #'(0.7 . 0)
                                \times 2/3 {
                                    c8
                                }
                            }
                        >>
                        \layout {
                            indent = #0
                            ragged-right = ##t
                        }
                    }
                }

    ..  container:: example

        With ties:

        >>> maker = abjad.NoteMaker()
        >>> notes = maker([0], [(5, 16)])
        >>> metric_modulation = abjad.MetricModulation(
        ...     left_rhythm=abjad.Note("c'4"),
        ...     right_rhythm=notes,
        ...     )

        >>> abjad.show(metric_modulation) # doctest: +SKIP

        ..  docs::

            >>> print(format(metric_modulation, 'lilypond'))
            \markup {
                \score
                    {
                        \new Score
                        \with
                        {
                            \override SpacingSpanner.spacing-increment = #0.5
                            proportionalNotationDuration = ##f
                        }
                        <<
                            \new RhythmicStaff
                            \with
                            {
                                \remove Time_signature_engraver
                                \remove Staff_symbol_engraver
                                \override Stem.direction = #up
                                \override Stem.length = #5
                                \override TupletBracket.bracket-visibility = ##t
                                \override TupletBracket.direction = #up
                                \override TupletBracket.minimum-length = #4
                                \override TupletBracket.padding = #1.25
                                \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                \override TupletNumber.font-size = #0
                                \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                tupletFullLength = ##t
                            }
                            {
                                c'4
                            }
                        >>
                        \layout {
                            indent = #0
                            ragged-right = ##t
                        }
                    }
                =
                \hspace
                    #-0.5
                \score
                    {
                        \new Score
                        \with
                        {
                            \override SpacingSpanner.spacing-increment = #0.5
                            proportionalNotationDuration = ##f
                        }
                        <<
                            \new RhythmicStaff
                            \with
                            {
                                \remove Time_signature_engraver
                                \remove Staff_symbol_engraver
                                \override Stem.direction = #up
                                \override Stem.length = #5
                                \override TupletBracket.bracket-visibility = ##t
                                \override TupletBracket.direction = #up
                                \override TupletBracket.minimum-length = #4
                                \override TupletBracket.padding = #1.25
                                \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                \override TupletNumber.font-size = #0
                                \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                tupletFullLength = ##t
                            }
                            {
                                c'4
                                ~
                                c'16
                            }
                        >>
                        \layout {
                            indent = #0
                            ragged-right = ##t
                        }
                    }
                }

    ..  container:: example

        With ties and tuplets:

        >>> maker = abjad.NoteMaker()
        >>> notes = maker([0], [(5, 16)])
        >>> tuplet = abjad.Tuplet((2, 3), notes)
        >>> metric_modulation = abjad.MetricModulation(
        ...     left_rhythm=abjad.Note("c'4"),
        ...     right_rhythm=tuplet,
        ...     )

        >>> abjad.show(metric_modulation) # doctest: +SKIP

        ..  docs::

            >>> print(format(metric_modulation, 'lilypond'))
            \markup {
                \score
                    {
                        \new Score
                        \with
                        {
                            \override SpacingSpanner.spacing-increment = #0.5
                            proportionalNotationDuration = ##f
                        }
                        <<
                            \new RhythmicStaff
                            \with
                            {
                                \remove Time_signature_engraver
                                \remove Staff_symbol_engraver
                                \override Stem.direction = #up
                                \override Stem.length = #5
                                \override TupletBracket.bracket-visibility = ##t
                                \override TupletBracket.direction = #up
                                \override TupletBracket.minimum-length = #4
                                \override TupletBracket.padding = #1.25
                                \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                \override TupletNumber.font-size = #0
                                \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                tupletFullLength = ##t
                            }
                            {
                                c'4
                            }
                        >>
                        \layout {
                            indent = #0
                            ragged-right = ##t
                        }
                    }
                =
                \hspace
                    #-0.5
                \score
                    {
                        \new Score
                        \with
                        {
                            \override SpacingSpanner.spacing-increment = #0.5
                            proportionalNotationDuration = ##f
                        }
                        <<
                            \new RhythmicStaff
                            \with
                            {
                                \remove Time_signature_engraver
                                \remove Staff_symbol_engraver
                                \override Stem.direction = #up
                                \override Stem.length = #5
                                \override TupletBracket.bracket-visibility = ##t
                                \override TupletBracket.direction = #up
                                \override TupletBracket.minimum-length = #4
                                \override TupletBracket.padding = #1.25
                                \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                \override TupletNumber.font-size = #0
                                \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                tupletFullLength = ##t
                            }
                            {
                                \tweak edge-height #'(0.7 . 0)
                                \times 2/3 {
                                    c'4
                                    ~
                                    c'16
                                }
                            }
                        >>
                        \layout {
                            indent = #0
                            ragged-right = ##t
                        }
                    }
                }


    ..  container:: example

        Attach metric modulations to generate score output:

        >>> staff = abjad.Staff("c'4 d'4 e'4 f'4 e'4 d'4")
        >>> abjad.attach(abjad.TimeSignature((3, 4)), staff[0])
        >>> score = abjad.Score([staff])

        >>> metric_modulation = abjad.MetricModulation(
        ...     left_rhythm=abjad.Note("c4"),
        ...     right_rhythm=abjad.Note("c8."),
        ...     )
        >>> abjad.attach(metric_modulation, staff[3])
        >>> abjad.override(staff).text_script.staff_padding = 2.5

        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(score)
            \new Score
            <<
                \new Staff
                \with
                {
                    \override TextScript.staff-padding = #2.5
                }
                {
                    \time 3/4
                    c'4
                    d'4
                    e'4
                    f'4
                    ^ \markup {
                        \score
                            {
                                \new Score
                                \with
                                {
                                    \override SpacingSpanner.spacing-increment = #0.5
                                    proportionalNotationDuration = ##f
                                }
                                <<
                                    \new RhythmicStaff
                                    \with
                                    {
                                        \remove Time_signature_engraver
                                        \remove Staff_symbol_engraver
                                        \override Stem.direction = #up
                                        \override Stem.length = #5
                                        \override TupletBracket.bracket-visibility = ##t
                                        \override TupletBracket.direction = #up
                                        \override TupletBracket.minimum-length = #4
                                        \override TupletBracket.padding = #1.25
                                        \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                        \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                        \override TupletNumber.font-size = #0
                                        \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                        tupletFullLength = ##t
                                    }
                                    {
                                        c4
                                    }
                                >>
                                \layout {
                                    indent = #0
                                    ragged-right = ##t
                                }
                            }
                        =
                        \hspace
                            #-0.5
                        \score
                            {
                                \new Score
                                \with
                                {
                                    \override SpacingSpanner.spacing-increment = #0.5
                                    proportionalNotationDuration = ##f
                                }
                                <<
                                    \new RhythmicStaff
                                    \with
                                    {
                                        \remove Time_signature_engraver
                                        \remove Staff_symbol_engraver
                                        \override Stem.direction = #up
                                        \override Stem.length = #5
                                        \override TupletBracket.bracket-visibility = ##t
                                        \override TupletBracket.direction = #up
                                        \override TupletBracket.minimum-length = #4
                                        \override TupletBracket.padding = #1.25
                                        \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                        \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                        \override TupletNumber.font-size = #0
                                        \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                        tupletFullLength = ##t
                                    }
                                    {
                                        c8.
                                    }
                                >>
                                \layout {
                                    indent = #0
                                    ragged-right = ##t
                                }
                            }
                        }
                    e'4
                    d'4
                }
            >>

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_hide',
        '_left_markup',
        '_left_rhythm',
        '_right_markup',
        '_right_rhythm',
        )

    _publish_storage_format = True

    ### INITIALIZER ###

    def __init__(
        self,
        left_rhythm=None,
        right_rhythm=None,
        *,
        hide: bool = None,
        left_markup: Markup = None,
        right_markup: Markup = None,
        ) -> None:
        from abjad.core.Note import Note
        if hide is not None:
            hide = bool(hide)
        self._hide = hide
        left_rhythm = left_rhythm or Note('c4')
        right_rhythm = right_rhythm or Note('c4')
        left_rhythm = self._initialize_rhythm(left_rhythm)
        self._left_rhythm = left_rhythm
        right_rhythm = self._initialize_rhythm(right_rhythm)
        self._right_rhythm = right_rhythm
        self._right_rhythm = right_rhythm
        if left_markup is not None:
            assert isinstance(left_markup, Markup)
        self._left_markup = left_markup
        if right_markup is not None:
            assert isinstance(right_markup, Markup)
        self._right_markup = right_markup

    ### SPECIAL METHODS ###

    def __eq__(self, argument) -> bool:
        """
        Is true when ``argument`` is another metric modulation with the same
        ratio as this metric modulation.

        ..  container:: example

            >>> metric_modulation_1 = abjad.MetricModulation(
            ...     left_rhythm=abjad.Note("c'4"),
            ...     right_rhythm=abjad.Note("c'4."),
            ...     )
            >>> metric_modulation_2 = abjad.MetricModulation(
            ...     left_rhythm=abjad.Tuplet((2, 3), [abjad.Note("c'4")]),
            ...     right_rhythm=abjad.Note("c'4"),
            ...     )
            >>> maker = abjad.NoteMaker()
            >>> notes = maker([0], [(5, 16)])
            >>> metric_modulation_3 = abjad.MetricModulation(
            ...     left_rhythm=abjad.Note("c'4"),
            ...     right_rhythm=notes,
            ...     )

            >>> metric_modulation_1.ratio
            Ratio((2, 3))
            >>> metric_modulation_2.ratio
            Ratio((2, 3))
            >>> metric_modulation_3.ratio
            Ratio((4, 5))

            >>> metric_modulation_1 == metric_modulation_1
            True
            >>> metric_modulation_1 == metric_modulation_2
            True
            >>> metric_modulation_1 == metric_modulation_3
            False

            >>> metric_modulation_2 == metric_modulation_1
            True
            >>> metric_modulation_2 == metric_modulation_2
            True
            >>> metric_modulation_2 == metric_modulation_3
            False

            >>> metric_modulation_3 == metric_modulation_1
            False
            >>> metric_modulation_3 == metric_modulation_2
            False
            >>> metric_modulation_3 == metric_modulation_3
            True

        """
        # custom definition because input rhythms don't compare:
        if isinstance(argument, type(self)):
            if self.ratio == argument.ratio:
                return True
        return False

    def __format__(self, format_specification='') -> str:
        """
        Formats metric modulation.

        ..  container:: example

            >>> metric_modulation = abjad.MetricModulation(
            ...     left_rhythm=abjad.Note("c'4"),
            ...     right_rhythm=abjad.Note("c'4."),
            ...     )

            >>> abjad.f(metric_modulation)
            abjad.MetricModulation(
                left_rhythm=abjad.Selection(
                    [
                        abjad.Note("c'4"),
                        ]
                    ),
                right_rhythm=abjad.Selection(
                    [
                        abjad.Note("c'4."),
                        ]
                    ),
                )

        """
        return super().__format__(
            format_specification=format_specification
            )

    def __hash__(self) -> int:
        """
        Hashes metric modulation.

        Redefined in tandem with __eq__.
        """
        return super().__hash__()

    def __illustrate__(self):
        r"""
        Illustrates metric modulation.

        ..  container:: example

            >>> metric_modulation = abjad.MetricModulation(
            ...     left_rhythm=abjad.Tuplet((2, 3), "c'4"),
            ...     right_rhythm=abjad.Note("c'4."),
            ...     )
            >>> abjad.show(metric_modulation) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = metric_modulation.__illustrate__()
                >>> metric_modulation = lilypond_file.items[-1]
                >>> print(format(metric_modulation))
                \markup {
                    \score
                        {
                            \new Score
                            \with
                            {
                                \override SpacingSpanner.spacing-increment = #0.5
                                proportionalNotationDuration = ##f
                            }
                            <<
                                \new RhythmicStaff
                                \with
                                {
                                    \remove Time_signature_engraver
                                    \remove Staff_symbol_engraver
                                    \override Stem.direction = #up
                                    \override Stem.length = #5
                                    \override TupletBracket.bracket-visibility = ##t
                                    \override TupletBracket.direction = #up
                                    \override TupletBracket.minimum-length = #4
                                    \override TupletBracket.padding = #1.25
                                    \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                    \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                    \override TupletNumber.font-size = #0
                                    \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                    tupletFullLength = ##t
                                }
                                {
                                    \tweak edge-height #'(0.7 . 0)
                                    \times 2/3 {
                                        c'4
                                    }
                                }
                            >>
                            \layout {
                                indent = #0
                                ragged-right = ##t
                            }
                        }
                    =
                    \hspace
                        #-0.5
                    \score
                        {
                            \new Score
                            \with
                            {
                                \override SpacingSpanner.spacing-increment = #0.5
                                proportionalNotationDuration = ##f
                            }
                            <<
                                \new RhythmicStaff
                                \with
                                {
                                    \remove Time_signature_engraver
                                    \remove Staff_symbol_engraver
                                    \override Stem.direction = #up
                                    \override Stem.length = #5
                                    \override TupletBracket.bracket-visibility = ##t
                                    \override TupletBracket.direction = #up
                                    \override TupletBracket.minimum-length = #4
                                    \override TupletBracket.padding = #1.25
                                    \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                    \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                    \override TupletNumber.font-size = #0
                                    \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                    tupletFullLength = ##t
                                }
                                {
                                    c'4.
                                }
                            >>
                            \layout {
                                indent = #0
                                ragged-right = ##t
                            }
                        }
                    }

        Returns LilyPond file.
        """
        import abjad
        lilypond_file = abjad.LilyPondFile.new()
        lilypond_file.items.append(self._get_markup())
        return lilypond_file

    def __str__(self) -> str:
        r"""
        Gets string representation of metric modulation.

        ..  container:: example

            >>> metric_modulation = abjad.MetricModulation(
            ...     left_rhythm=abjad.Tuplet((2, 3), [abjad.Note("c'4")]),
            ...     right_rhythm=abjad.Note("c'4"),
            ...     )

            >>> print(str(metric_modulation))
            \markup {
                \score
                    {
                        \new Score
                        \with
                        {
                            \override SpacingSpanner.spacing-increment = #0.5
                            proportionalNotationDuration = ##f
                        }
                        <<
                            \new RhythmicStaff
                            \with
                            {
                                \remove Time_signature_engraver
                                \remove Staff_symbol_engraver
                                \override Stem.direction = #up
                                \override Stem.length = #5
                                \override TupletBracket.bracket-visibility = ##t
                                \override TupletBracket.direction = #up
                                \override TupletBracket.minimum-length = #4
                                \override TupletBracket.padding = #1.25
                                \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                \override TupletNumber.font-size = #0
                                \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                tupletFullLength = ##t
                            }
                            {
                                \tweak edge-height #'(0.7 . 0)
                                \times 2/3 {
                                    c'4
                                }
                            }
                        >>
                        \layout {
                            indent = #0
                            ragged-right = ##t
                        }
                    }
                =
                \hspace
                    #-0.5
                \score
                    {
                        \new Score
                        \with
                        {
                            \override SpacingSpanner.spacing-increment = #0.5
                            proportionalNotationDuration = ##f
                        }
                        <<
                            \new RhythmicStaff
                            \with
                            {
                                \remove Time_signature_engraver
                                \remove Staff_symbol_engraver
                                \override Stem.direction = #up
                                \override Stem.length = #5
                                \override TupletBracket.bracket-visibility = ##t
                                \override TupletBracket.direction = #up
                                \override TupletBracket.minimum-length = #4
                                \override TupletBracket.padding = #1.25
                                \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                \override TupletNumber.font-size = #0
                                \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                tupletFullLength = ##t
                            }
                            {
                                c'4
                            }
                        >>
                        \layout {
                            indent = #0
                            ragged-right = ##t
                        }
                    }
                }

        """
        return str(self._get_markup())

    ### PRIVATE METHODS ###

    def _get_left_markup(self):
        if self.left_markup is not None:
            return self.left_markup
        markup = Duration._to_score_markup(self.left_rhythm)
        return markup

    def _get_lilypond_format(self):
        return str(self)

    def _get_lilypond_format_bundle(self, component=None):
        bundle = LilyPondFormatBundle()
        if not self.hide:
            markup = self._get_markup()
            markup = new(markup, direction=enums.Up)
            markup_format_pieces = markup._get_format_pieces()
            bundle.after.markup.extend(markup_format_pieces)
        return bundle

    def _get_markup(self):
        left_markup = self._get_left_markup()
        equal = Markup('=')
        right_space = Markup.hspace(-0.5)
        right_markup = self._get_right_markup()
        markup = left_markup + equal + right_space + right_markup
        return markup

    def _get_right_markup(self):
        if self.right_markup is not None:
            return self.right_markup
        markup = Duration._to_score_markup(self.right_rhythm)
        return markup

    def _initialize_rhythm(self, rhythm):
        import abjad
        if isinstance(rhythm, abjad.Component):
            selection = select([rhythm])
        elif isinstance(rhythm, abjad.Selection):
            selection = rhythm
        else:
            message = 'rhythm must be duration, component or selection: {!r}.'
            message = message.format(rhythm)
            raise TypeError(message)
        assert isinstance(selection, abjad.Selection)
        return selection

    ### PUBLIC PROPERTIES ###

    @property
    def hide(self) -> typing.Optional[bool]:
        """
        Is true when metric modulation generates no LilyPond output.
        """
        return self._hide

    @property
    def left_markup(self) -> typing.Optional[Markup]:
        """
        Gets left markup of metric modulation.

        ..  container:: example

            >>> metric_modulation = abjad.MetricModulation(
            ...     left_rhythm=abjad.Note("c'4"),
            ...     right_rhythm=abjad.Note("c'4."),
            ...     )
            >>> metric_modulation.left_markup

        """
        return self._left_markup

    @property
    def left_rhythm(self):
        """
        Gets left rhythm of metric modulation.

        ..  container:: example

            >>> metric_modulation = abjad.MetricModulation(
            ...     left_rhythm=abjad.Note("c'4"),
            ...     right_rhythm=abjad.Note("c'4."),
            ...     )
            >>> metric_modulation.left_rhythm
            Selection([Note("c'4")])

        Returns selection.
        """
        return self._left_rhythm

    @property
    def ratio(self) -> Ratio:
        """
        Gets ratio of metric modulation.

        ..  container:: example

            >>> metric_modulation = abjad.MetricModulation(
            ...     left_rhythm=abjad.Tuplet((2, 3), [abjad.Note("c'4")]),
            ...     right_rhythm=abjad.Note("c'4"),
            ...     )
            >>> metric_modulation.ratio
            Ratio((2, 3))

        """
        left_duration = inspect(self.left_rhythm).duration()
        right_duration = inspect(self.right_rhythm).duration()
        duration = left_duration / right_duration
        ratio = Ratio(duration.pair)
        return ratio

    @property
    def right_markup(self) -> typing.Optional[Markup]:
        r"""Gets right markup of metric modulation.

        ..  container:: example

            >>> metric_modulation = abjad.MetricModulation(
            ...     left_rhythm=abjad.Note("c'4"),
            ...     right_rhythm=abjad.Note("c'4."),
            ...     )
            >>> metric_modulation.right_markup

        """
        return self._right_markup

    @property
    def right_rhythm(self):
        """
        Gets right tempo of metric modulation.

        ..  container:: example

            >>> metric_modulation = abjad.MetricModulation(
            ...     left_rhythm=abjad.Note("c'4"),
            ...     right_rhythm=abjad.Note("c'4."),
            ...     )
            >>> metric_modulation.right_rhythm
            Selection([Note("c'4.")])

        """
        return self._right_rhythm

    @property
    def tweaks(self) -> None:
        """
        Are not implemented on metric modulation.
        """
        pass



"""
Abjad's top-level functions.
"""

from .activate import activate
from .annotate import annotate
from .attach import attach
from .deactivate import deactivate
from .detach import detach
from .f import f
from .graph import graph
from .inspect import inspect
from .iterate import iterate
from .label import label
from .mutate import mutate
from .new import new
from .override import override
from .parse import parse
from .persist import persist
from .play import play
from .select import select
from .sequence import sequence
from .setting import setting
from .show import show
from .tweak import tweak



def setting(argument):
    r"""
    Makes LilyPond setting name manager.

    ..  container:: example

        Sets instrument name:

        >>> staff = abjad.Staff("c'4 e'4 d'4 f'4")
        >>> abjad.setting(staff).instrument_name = abjad.Markup('Vn. I')
        >>> abjad.show(staff) # doctest: +SKIP


        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            \with
            {
                instrumentName = \markup { "Vn. I" }
            }
            {
                c'4
                e'4
                d'4
                f'4
            }

    ..  container:: example

        Returns LilyPond setting name manager:

        >>> abjad.setting(staff)
        LilyPondSettingNameManager(('instrument_name', Markup(contents=['Vn. I'])))

    """
    import abjad
    if getattr(argument, '_lilypond_setting_name_manager', None) is None:
        manager = abjad.lilypondnames.LilyPondSettingNameManager()
        argument._lilypond_setting_name_manager = manager
    return argument._lilypond_setting_name_manager



import typing
from abjad import enums
from abjad.system.AbjadValueObject import AbjadValueObject
from abjad.lilypondnames.LilyPondTweakManager import LilyPondTweakManager
from abjad.system.LilyPondFormatBundle import LilyPondFormatBundle


class LaissezVibrer(AbjadValueObject):
    r"""
    Laissez vibrer.

    ..  container:: example

        >>> chord = abjad.Chord("<c' e' g' c''>4")
        >>> laissez_vibrer = abjad.LaissezVibrer()
        >>> abjad.attach(laissez_vibrer, chord)
        >>> abjad.show(chord) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(chord)
            <c' e' g' c''>4
            \laissezVibrer

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_tweaks',
        )

    _format_slot = 'after'

    _time_orientation: enums.HorizontalAlignment = enums.Right

    ### INITIALIZER ###

    def __init__(
        self,
        *,
        tweaks: typing.Union[
            typing.List[typing.Tuple], LilyPondTweakManager] = None,
        ) -> None:
        self._tweaks = None
        LilyPondTweakManager.set_tweaks(self, tweaks)

    ### SPECIAL METHODS ###

    def __str__(self) -> str:
        r"""
        Gets string representation of laissez vibrer indicator.

        ..  container:: example

            Default:

            >>> str(abjad.LaissezVibrer())
            '\\laissezVibrer'

        """
        return r'\laissezVibrer'

    ### PRIVATE METHODS ###

    def _get_lilypond_format(self):
        return str(self)

    def _get_lilypond_format_bundle(self, component=None):
        bundle = LilyPondFormatBundle()
        if self.tweaks:
            tweaks = self.tweaks._list_format_contributions()
            bundle.after.articulations.extend(tweaks)
        bundle.after.articulations.append(self._get_lilypond_format())
        return bundle

    ### PUBLIC PROPERTIES ###

    @property
    def tweaks(self) -> typing.Optional[LilyPondTweakManager]:
        r"""
        Gets tweaks

        ..  container:: example

            >>> note = abjad.Note("c'4")
            >>> lv = abjad.LaissezVibrer()
            >>> abjad.tweak(lv).color = 'blue'
            >>> abjad.attach(lv, note)
            >>> abjad.show(note) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(note)
                c'4
                - \tweak color #blue
                \laissezVibrer

        ..  container:: example

            >>> note = abjad.Note("c'4")
            >>> lv = abjad.LaissezVibrer(tweaks=[('color', 'blue')])
            >>> abjad.attach(lv, note)
            >>> abjad.show(note) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(note)
                c'4
                - \tweak color #blue
                \laissezVibrer

        """
        return self._tweaks



from typing import Callable, Dict  # noqa
from abjad import core
from abjad import indicators as abjad_indicators
from abjad import pitch as abjad_pitch
from abjad.system.AbjadObject import AbjadObject
from abjad.top import attach


# TODO: should not inherit from AbjadObject because no slots
class GuileProxy(AbjadObject):
    """
    Emulates LilyPond music functions.

    Used internally by LilyPondParser.

    Not composer-safe.
    """

    ### CLASS VARIABLES ###

    _function_name_mapping = {}  # type: Dict[str, Callable]

    ### INITIALIZER ###

    def __init__(self, client=None):
        self.client = client

    ### SPECIAL METHODS ###

    def __call__(self, function_name, arguments):
        """
        Calls Guile proxy on ``function_name`` with ``arguments``.

        Returns function output.
        """
        if hasattr(self, function_name[1:]):
            result = getattr(self, function_name[1:])(*arguments)
            return result
        elif function_name[1:] in self._function_name_mapping:
            function_name = function_name[1:]
            result = getattr(self, function_name)(*arguments)
            return result
        message = 'LilyPondParser can not emulate music function: {}.'
        message = message.format(function_name)
        raise Exception(message)

    ### FUNCTION EMULATORS ###

    def acciaccatura(self, music):
        r"""
        Handles LilyPond ``\acciaccatura`` command.
        """
        grace = core.AcciaccaturaContainer(music[:])
        return grace

    # afterGrace?

    def appoggiatura(self, music):
        r"""
        Handles LilyPond ``\appoggiatura`` command.
        """
        grace = core.AppoggiaturaContainer(music[:])
        return grace

    def bar(self, string):
        r"""
        Handles LilyPond ``\bar`` command.
        """
        return abjad_indicators.BarLine(string)

    def breathe(self):
        r"""
        Handles LilyPond ``\breathe`` command.
        """
        return abjad_indicators.LilyPondLiteral(r'\breathe', 'after')

    def clef(self, string):
        r"""
        Handles LilyPond ``\clef`` command.
        """
        return abjad_indicators.Clef(string)

    def grace(self, music):
        r"""
        Handles LilyPond ``\grace`` command.
        """
        assert isinstance(music, core.Container)
        leaves = music[:]
        music[:] = []
        return core.GraceContainer(leaves)

    def key(self, notename_pitch, number_list):
        r"""
        Handles LilyPond ``\key`` command.
        """
        if number_list is None:
            number_list = 'major'
        return abjad_indicators.KeySignature(notename_pitch, number_list)

    def language(self, string):
        r"""
        Handles LilyPond ``\language`` command.
        """
        if string in self.client._language_pitch_names:
            self.client._pitch_names = \
                self.client._language_pitch_names[string]
        # try reparsing the next note name, if a note name immediately follows
        lookahead = self.client._parser.lookahead
        if lookahead.type == 'STRING':
            if lookahead.value in self.client._pitch_names:
                lookahead.type = 'NOTENAME_PITCH'
                lookahead.value = self.client._pitch_names[lookahead.value]

    def makeClusters(self, music):
        r"""
        Handles LilyPond ``\makeClusters`` command.
        """
        return core.Cluster(music[:])

    def mark(self, label):
        r"""
        Handles LilyPond ``\mark`` command.
        """
        if label is None:
            label = '\default'
        return abjad_indicators.LilyPondLiteral(r'\mark %s' % label)

    def oneVoice(self):
        r"""
        Handles LilyPond ``\oneVoice`` command.
        """
        return abjad_indicators.LilyPondLiteral(r'\oneVoice')

    # pitchedTrill

    def relative(self, pitch, music):
        r"""
        Handles LilyPond ``\relative`` command.
        """
        # We should always keep track of the last chord entered.
        # When there are repeated chords (via q),
        # we add the last chord as a key in a _repeated_chords dictionary.
        # Then, we associate a list with the chord "key" in the dict,
        # and append a reference to the repeated chord.

        # Should the referenced chord appear in a relative block,
        # we relativize that chord, and update any repeated chords
        # we've added to its list of referencing chords.

        # The parser's "last_chord" variable will now reflect the
        # relativized pitches of the original referenced chord,
        # and so any new chord repetitions following the \relative block
        # should result in matching absolute pitches to both the "last_chord"
        # and any other repetitions.

        if self._is_unrelativable(music):
            return music

        def recurse(component, pitch):
            if self._is_unrelativable(component):
                return pitch
            elif isinstance(component, (core.Chord, core.Note)):
                pitch = self._make_relative_leaf(component, pitch)
                if component in self.client._repeated_chords:
                    for repeated_chord in \
                        self.client._repeated_chords[component]:
                        repeated_chord.written_pitches = \
                            component.written_pitches
            elif isinstance(component, core.Container):
                for child in component:
                    pitch = recurse(child, pitch)
            return pitch

        pitch = recurse(music, pitch)

        self._make_unrelativable(music)

        return music

    def skip(self, duration):
        r"""
        Handles LilyPond ``\skip`` command.
        """
        leaf = core.Skip(duration.duration)
        if duration.multiplier is not None:
            attach(duration.multiplier, leaf)
        return leaf

    def slashed_grace_container(self, music):
        r"""
        Handles LilyPond ``\slahsedGrace`` command.
        """
        grace = core.GraceContainer(music[:])
        return grace

    def time(self, number_list, fraction):
        r"""
        Handles LilyPond ``\time`` command.
        """
        n, d = fraction.numerator, fraction.denominator
        return abjad_indicators.TimeSignature((n, d))

    def times(self, fraction, music):
        r"""
        Handles LilyPond ``\times`` command.
        """
        n, d = fraction.numerator, fraction.denominator
        if (not isinstance(music, core.Context) and
            not isinstance(music, core.Leaf)
            ):
            assert isinstance(music, core.Container), repr(music)
            leaves = music[:]
            music[:] = []
            return core.Tuplet((n, d), leaves)
        return core.Tuplet((n, d), [music])

    def transpose(self, from_pitch, to_pitch, music):
        r"""
        Handles LilyPond ``\transpose`` command.
        """
        from abjad import parser as abjad_parser
        def recurse(music):
            key_signatures = music._get_indicators(abjad_indicators.KeySignature)
            if key_signatures:
                for x in key_signatures:
                    tonic = abjad_pitch.NamedPitch((x.tonic.name, 4))
                    # TODO: cheating to assign to a read-only property
                    x._tonic = abjad_parser.LilyPondParser._transpose_enharmonically(
                        from_pitch, to_pitch, tonic).pitch_class
            if isinstance(music, core.Note):
                music.written_pitch = \
                    abjad_parser.LilyPondParser._transpose_enharmonically(
                        from_pitch, to_pitch, music.written_pitch)
            elif isinstance(music, core.Chord):
                for note_head in music.note_heads:
                    note_head.written_pitch = \
                        abjad_parser.LilyPondParser._transpose_enharmonically(
                            from_pitch, to_pitch, note_head.written_pitch)
            elif isinstance(music, core.Container):
                for x in music:
                    recurse(x)
        self._make_unrelativable(music)
        recurse(music)
        return music

    # transposition

    # tweak

    def voiceFour(self):
        r"""
        Handles LilyPond ``\voiceFour`` command.
        """
        return abjad_indicators.LilyPondLiteral(r'\voiceFour')

    def voiceOne(self):
        r"""
        Handles LilyPond ``\voiceOnce`` command.
        """
        return abjad_indicators.LilyPondLiteral(r'\voiceOne')

    def voiceThree(self):
        r"""
        Handles LilyPond ``\voiceThree`` command.
        """
        return abjad_indicators.LilyPondLiteral(r'\voiceThree')

    def voiceTwo(self):
        r"""
        Handles LilyPond ``\voiceTwo`` command.
        """
        return abjad_indicators.LilyPondLiteral(r'\voiceTwo')

    ### HELPER FUNCTIONS ###

    def _is_unrelativable(self, music):
        annotations = music._get_indicators(dict)
        keys = [list(_.keys())[0] for _ in annotations]
        if 'UnrelativableMusic' in keys:
            return True
        return False

    def _make_relative_leaf(self, leaf, pitch):
        if self._is_unrelativable(leaf):
            return pitch
        elif isinstance(leaf, core.Note):
            pitch = self._to_relative_octave(leaf.written_pitch, pitch)
            leaf.written_pitch = pitch
        elif isinstance(leaf, core.Chord):
            # TODO: This is not ideal w/r/t post events as LilyPond does
            # not sort chord contents
            chord_pitches = self.client._chord_pitch_orders[leaf]
            for i, chord_pitch in enumerate(chord_pitches):
                pitch = self._to_relative_octave(chord_pitch, pitch)
                chord_pitches[i] = pitch
            leaf.written_pitches = chord_pitches
            pitch = min(leaf.written_pitches)
        return pitch

    def _make_unrelativable(self, music):
        if not self._is_unrelativable(music):
            annotation = {'UnrelativableMusic': True}
            attach(annotation, music)

    def _to_relative_octave(self, pitch, reference):
        if pitch.pitch_class.number > reference.pitch_class.number:
            pair = (pitch.pitch_class.name, reference.octave.number)
            up_pitch = abjad_pitch.NamedPitch(pair)
            pair = (pitch.pitch_class.name, reference.octave.number - 1)
            down_pitch = abjad_pitch.NamedPitch(pair)
            up_octave = up_pitch.octave.number
            down_octave = down_pitch.octave.number
        else:
            pair = (pitch.pitch_class.name, reference.octave.number + 1)
            up_pitch = abjad_pitch.NamedPitch(pair)
            pair = (pitch.pitch_class.name, reference.octave.number)
            down_pitch = abjad_pitch.NamedPitch(pair)
            up_octave = up_pitch.octave.number
            down_octave = down_pitch.octave.number
        if abs(
                float(up_pitch._get_diatonic_pitch_number()) -
                float(reference._get_diatonic_pitch_number())) < \
            abs(
                float(down_pitch._get_diatonic_pitch_number()) -
                float(reference._get_diatonic_pitch_number())):
            pair = (
                up_pitch.pitch_class.name,
                up_octave + pitch.octave.number - 3,
                )
            pitch = abjad_pitch.NamedPitch(pair)
        else:
            pair = (
                down_pitch.pitch_class.name,
                down_octave + pitch.octave.number - 3,
                )
            pitch = abjad_pitch.NamedPitch(pair)
        return pitch



from fractions import Fraction  # noqa
try:
    from quicktions import Fraction  # type: ignore
except ImportError:
    pass

from abjad.enums import *
from abjad.exceptions import *

# ensure that the ~/.abjad directory and friends are setup
# and instantiate Abjad's configuration singleton
from abjad.system.AbjadConfiguration import AbjadConfiguration
abjad_configuration = AbjadConfiguration()
del AbjadConfiguration

# import parser
from abjad import parser

# import all tools packages
from abjad.system import *
from abjad.utilities import *

# typings after utilities (for Expression)
from abjad.typings import *

index = Pattern.index
index_all = Pattern.index_all
index_first = Pattern.index_first
index_last = Pattern.index_last

from abjad.indicators import *
from abjad.instruments import *
from abjad.lilypondfile import *
from abjad.lilypondnames import *
from abjad.markups import *
from abjad.meter import *
from abjad.pitch import *
from abjad.scheme import *
from abjad.core import *
from abjad.segments import *
from abjad.spanners import *
from abjad.top import *

# import all the way down to module to satisfy mypy:
from abjad.mathtools.NonreducedFraction import NonreducedFraction
from abjad.mathtools.NonreducedRatio import NonreducedRatio
from abjad.mathtools.Ratio import Ratio

# timespan classes (but not functions):
from abjad.timespans.AnnotatedTimespan import AnnotatedTimespan
from abjad.timespans.Timespan import Timespan
from abjad.timespans.TimespanInequality import TimespanInequality
from abjad.timespans.TimespanList import TimespanList

# import version information
from ._version import __version_info__, __version__
try:
    del _version
except NameError:
    pass

from abjad import cli
from abjad import demos
from abjad import ly
from abjad import utilities

# singletons
tags = system.Tags()



import copy
import typing
from abjad.system.AbjadValueObject import AbjadValueObject
from abjad.lilypondnames.LilyPondTweakManager import LilyPondTweakManager
from abjad.markups import Markup
from abjad.system.LilyPondFormatBundle import LilyPondFormatBundle
from abjad.top.new import new


class RehearsalMark(AbjadValueObject):
    r"""
    Rehearsal mark.

    ..  container:: example

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> score = abjad.Score([staff])
        >>> mark = abjad.RehearsalMark(number=1)
        >>> abjad.attach(mark, staff[0])
        >>> scheme = abjad.Scheme('format-mark-box-alphabet')
        >>> abjad.setting(score).markFormatter = scheme
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(score)
            \new Score
            \with
            {
                markFormatter = #format-mark-box-alphabet
            }
            <<
                \new Staff
                {
                    \mark #1
                    c'4
                    d'4
                    e'4
                    f'4
                }
            >>

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_markup',
        '_number',
        '_tweaks',
        )

    _context = 'Score'

    ### INITIALIZER ###

    def __init__(
        self,
        *,
        number: int = None,
        markup: Markup = None,
        tweaks: typing.Union[
            typing.List[typing.Tuple], LilyPondTweakManager] = None,
        ) -> None:
        self._tweaks = None
        self._markup = markup
        self._number = number
        LilyPondTweakManager.set_tweaks(self, tweaks)

    ### SPECIAL METHODS ###

    def __copy__(self, *arguments):
        r"""
        Copies rehearsal mark.

        >>> import copy

        ..  container:: example

            Preserves tweaks:

            >>> mark = abjad.RehearsalMark(number=1)
            >>> abjad.tweak(mark).color = 'red'
            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> abjad.attach(mark, staff[0])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \tweak color #red
                    \mark #1
                    c'4
                    d'4
                    e'4
                    f'4
                }

            >>> mark = copy.copy(mark)
            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> abjad.attach(mark, staff[0])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \tweak color #red
                    \mark #1
                    c'4
                    d'4
                    e'4
                    f'4
                }

        """
        mark_ = new(self)
        mark_._tweaks = copy.copy(self.tweaks)
        return mark_

    def __str__(self) -> str:
        r"""
        Gets string representation of rehearsal mark.

        ..  container:: example

            >>> mark = abjad.RehearsalMark(number=1)
            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> abjad.attach(mark, staff[0])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::
            
                >>> abjad.f(staff)
                \new Staff
                {
                    \mark #1
                    c'4
                    d'4
                    e'4
                    f'4
                }

            >>> str(mark)
            '\\mark #1'

        """
        return self._get_lilypond_format()

    ### PRIVATE METHODS ###

    def _get_lilypond_format(self):
        if self.markup is not None:
            result = rf'\mark {self.markup}'
        elif self.number is not None:
            result = rf'\mark #{self.number}'
        else:
            result = r'\mark \default'
        return result

    def _get_lilypond_format_bundle(self, component=None):
        bundle = LilyPondFormatBundle()
        if self.tweaks:
            tweaks = self.tweaks._list_format_contributions(directed=False)
            bundle.opening.commands.extend(tweaks)
        bundle.opening.commands.append(self._get_lilypond_format())
        return bundle

    ### PUBLIC PROPERTIES ###

    @property
    def context(self) -> str:
        """
        Is ``'Score'``.

        ..  container:: example

            >>> abjad.RehearsalMark(number=1).context
            'Score'

        """
        return self._context

    @property
    def markup(self) -> typing.Optional[Markup]:
        r"""
        Gets rehearsal mark markup.

        ..  container:: example

            >>> markup = abjad.Markup(r'\bold { \italic { A } }')
            >>> mark = abjad.RehearsalMark(markup=markup)
            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> abjad.attach(mark, staff[0])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \mark \markup {
                        \bold
                            {
                                \italic
                                    {
                                        A
                                    }
                            }
                        }
                    c'4
                    d'4
                    e'4
                    f'4
                }

        """
        return self._markup

    @property
    def number(self) -> typing.Optional[int]:
        """
        Gets rehearsal mark number.

        ..  container:: example

            >>> abjad.RehearsalMark(number=1).number
            1

        """
        return self._number

    @property
    def tweaks(self) -> typing.Optional[LilyPondTweakManager]:
        r"""
        Gets tweaks

        ..  container:: example

            >>> note = abjad.Note("c'4")
            >>> mark = abjad.RehearsalMark(markup='A')
            >>> abjad.tweak(mark).color = 'blue'
            >>> abjad.attach(mark, note)
            >>> abjad.show(note) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(note)
                \tweak color #blue
                \mark A
                c'4

        ..  container:: example

            >>> note = abjad.Note("c'4")
            >>> mark = abjad.RehearsalMark(
            ...     markup='A',
            ...     tweaks=[('color', 'blue')],
            ...     )
            >>> abjad.attach(mark, note)
            >>> abjad.show(note) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(note)
                \tweak color #blue
                \mark A
                c'4

        """
        return self._tweaks

    ### PUBLIC METHODS ###

    @staticmethod
    def from_string(string) -> 'RehearsalMark':
        """
        Makes rehearsal mark from ``string``.

        ..  container:: example

            >>> abjad.RehearsalMark.from_string('A')
            RehearsalMark(number=1)

        ..  container:: example

            >>> abjad.RehearsalMark.from_string('AA')
            RehearsalMark(number=27)

        ..  container:: example

            >>> abjad.RehearsalMark.from_string('AB')
            RehearsalMark(number=28)

        ..  container:: example

            >>> abjad.RehearsalMark.from_string('BA')
            RehearsalMark(number=53)

        ..  container:: example

            >>> abjad.RehearsalMark.from_string('BB')
            RehearsalMark(number=54)

        """
        number = 0
        for place, letter in enumerate(reversed(string)):
            integer = ord(letter) - ord('A') + 1
            multiplier = 26 ** place 
            integer *= multiplier
            number += integer
        return RehearsalMark(number=number)



import typing
from abjad.system.AbjadValueObject import AbjadValueObject
from abjad.lilypondnames.LilyPondTweakManager import LilyPondTweakManager
from abjad.pitch.NamedPitchClass import NamedPitchClass
from abjad.system.FormatSpecification import FormatSpecification
from abjad.system.LilyPondFormatBundle import LilyPondFormatBundle
from .Mode import Mode


class KeySignature(AbjadValueObject):
    r"""
    Key signature.

    ..  container:: example

        >>> staff = abjad.Staff("e'8 fs'8 gs'8 a'8")
        >>> key_signature = abjad.KeySignature('e', 'major')
        >>> abjad.attach(key_signature, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                \key e \major
                e'8
                fs'8
                gs'8
                a'8
            }

    ..  container:: example

        >>> staff = abjad.Staff("e'8 fs'8 g'8 a'8")
        >>> key_signature = abjad.KeySignature('e', 'minor')
        >>> abjad.attach(key_signature, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                \key e \minor
                e'8
                fs'8
                g'8
                a'8
            }

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_mode',
        '_tonic',
        '_tweaks',
        )

    _context = 'Staff'

    _format_slot = 'opening'

    _parameter = True
    
    _redraw = True

    ### INITIALIZER ###

    def __init__(
        self,
        tonic: str = 'c',
        mode: str = 'major',
        *,
        tweaks: typing.Union[
            typing.List[typing.Tuple], LilyPondTweakManager] = None,
        ) -> None:
        self._tonic = NamedPitchClass(tonic)
        self._mode = Mode(mode)
        self._tweaks = None
        LilyPondTweakManager.set_tweaks(self, tweaks)

    ### SPECIAL METHODS ###

    def __str__(self) -> str:
        """
        Gets string representation of key signature.

        ..  container:: example

            E major:

            >>> str(abjad.KeySignature('e', 'major'))
            'e-major'

        ..  container:: example

            e minor:

            >>> str(abjad.KeySignature('e', 'minor'))
            'e-minor'

        """
        return f'{self.tonic!s}-{self.mode!s}'

    ### PRIVATE METHODS ###

    def _get_format_specification(self):
        values = [self.tonic, self.mode]
        return FormatSpecification(
            client=self,
            repr_is_indented=False,
            storage_format_is_indented=False,
            storage_format_args_values=values,
            )

    def _get_lilypond_format(self):
        return rf'\key {self.tonic!s} \{self.mode!s}'

    def _get_lilypond_format_bundle(self, component=None):
        bundle = LilyPondFormatBundle()
        if self.tweaks:
            tweaks = self.tweaks._list_format_contributions(directed=False)
            bundle.before.commands.extend(tweaks)
        bundle.before.commands.append(self._get_lilypond_format())
        return bundle

    ### PUBLIC PROPERTIES ###

    @property
    def context(self) -> str:
        """
        Gets (historically conventional) context.

        ..  container:: example

            E major:

            >>> key_signature = abjad.KeySignature('e', 'major')
            >>> key_signature.context
            'Staff'

        ..  container:: example

            e minor:

            >>> key_signature = abjad.KeySignature('e', 'minor')
            >>> key_signature.context
            'Staff'

        Override with ``abjad.attach(..., context='...')``.
        """
        return self._context

    @property
    def mode(self):
        """
        Gets mode of key signature.

        ..  container:: example

            E major:

            >>> key_signature = abjad.KeySignature('e', 'major')
            >>> key_signature.mode
            Mode('major')

        ..  container:: example

            e minor:

            >>> key_signature = abjad.KeySignature('e', 'minor')
            >>> key_signature.mode
            Mode('minor')

        Returns mode.
        """
        return self._mode

    @property
    def name(self) -> str:
        """
        Gets name of key signature.

        ..  container:: example

            E major:

            >>> key_signature = abjad.KeySignature('e', 'major')
            >>> key_signature.name
            'E major'

        ..  container:: example

            e minor:

            >>> key_signature = abjad.KeySignature('e', 'minor')
            >>> key_signature.name
            'e minor'

        """
        if self.mode.mode_name == 'major':
            tonic = str(self.tonic).upper()
        else:
            tonic = str(self.tonic)
        return f'{tonic!s} {self.mode.mode_name!s}'

    @property
    def parameter(self) -> bool:
        """
        Is true.

        ..  container:: example

            >>> abjad.KeySignature('e', 'major').parameter
            True

        Class constant.
        """
        return self._parameter

    @property
    def redraw(self) -> bool:
        """
        Is true.

        ..  container:: example

            >>> abjad.KeySignature('e', 'major').redraw
            True

        Class constant.
        """
        return self._redraw

    @property
    def tonic(self) -> NamedPitchClass:
        """
        Gets tonic of key signature.

        ..  container:: example

            E major:

            >>> key_signature = abjad.KeySignature('e', 'major')
            >>> key_signature.tonic
            NamedPitchClass('e')

        ..  container:: example

            e minor:

            >>> key_signature = abjad.KeySignature('e', 'minor')
            >>> key_signature.tonic
            NamedPitchClass('e')

        """
        return self._tonic

    @property
    def tweaks(self) -> typing.Optional[LilyPondTweakManager]:
        r"""
        Gets tweaks.

        ..  container:: example

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> key = abjad.KeySignature('e', 'minor')
            >>> abjad.tweak(key).color = 'blue'
            >>> abjad.attach(key, staff[0])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \tweak color #blue
                    \key e \minor
                    c'4
                    d'4
                    e'4
                    f'4
                }

        ..  container:: example

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> key = abjad.KeySignature('e', 'minor', tweaks=[('color', 'blue')])
            >>> abjad.attach(key, staff[0])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \tweak color #blue
                    \key e \minor
                    c'4
                    d'4
                    e'4
                    f'4
                }

        """
        return self._tweaks



from .TypedTuple import TypedTuple


class PatternTuple(TypedTuple):
    """
    Pattern tuple.

    ..  container:: example

        Three patterns:

        >>> patterns = abjad.PatternTuple([
        ...     abjad.Pattern(
        ...         indices=[0, 1, 7],
        ...         period=10,
        ...         ),
        ...     abjad.Pattern(
        ...         indices=[-2, -1],
        ...         ),
        ...     abjad.Pattern(
        ...         indices=[2],
        ...         period=3,
        ...         ),
        ...     ])

        >>> abjad.f(patterns)
        abjad.PatternTuple(
            (
                abjad.Pattern(
                    indices=[0, 1, 7],
                    period=10,
                    ),
                abjad.Pattern(
                    indices=[-2, -1],
                    ),
                abjad.Pattern(
                    indices=[2],
                    period=3,
                    ),
                )
            )

    ..  container:: example

        Two patterns:

        >>> patterns = abjad.PatternTuple([
        ...     abjad.Pattern(
        ...         indices=[1],
        ...         period=2,
        ...         ),
        ...     abjad.Pattern(
        ...         indices=[-3, -2, -1],
        ...         ),
        ...     ])

        >>> abjad.f(patterns)
        abjad.PatternTuple(
            (
                abjad.Pattern(
                    indices=[1],
                    period=2,
                    ),
                abjad.Pattern(
                    indices=[-3, -2, -1],
                    ),
                )
            )

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        )

    ### PUBLIC METHODS ###

    def get_matching_pattern(self, index, total_length, rotation=None):
        """
        Gets pattern matching ``index``.

        ..  container:: example

            Two patterns:

            >>> patterns = abjad.PatternTuple([
            ...     abjad.Pattern(
            ...         indices=[1],
            ...         period=2,
            ...         ),
            ...     abjad.Pattern(
            ...         indices=[-3, -2, -1],
            ...         ),
            ...     ])

            Gets patterns that match the first ten indices:

            >>> for i in range(10):
            ...     match = patterns.get_matching_pattern(i, 10)
            ...     print(i, match)
            ...
            0 None
            1 Pattern(indices=[1], period=2)
            2 None
            3 Pattern(indices=[1], period=2)
            4 None
            5 Pattern(indices=[1], period=2)
            6 None
            7 Pattern(indices=[-3, -2, -1])
            8 Pattern(indices=[-3, -2, -1])
            9 Pattern(indices=[-3, -2, -1])

            Last three indices match the second pattern.

            Gets patterns that match next ten indices:

            >>> for i in range(10, 20):
            ...     match = patterns.get_matching_pattern(i, 10)
            ...     print(i, match)
            ...
            10 None
            11 Pattern(indices=[1], period=2)
            12 None
            13 Pattern(indices=[1], period=2)
            14 None
            15 Pattern(indices=[1], period=2)
            16 None
            17 Pattern(indices=[1], period=2)
            18 None
            19 Pattern(indices=[1], period=2)

            Last three indices no longer match the second pattern.

        ..  container:: example

            Gets patterns that match the first ten indices, with rotation set
            to ``1``:

            >>> for i in range(10):
            ...     match = patterns.get_matching_pattern(i, 10, rotation=1)
            ...     print(i, match)
            ...
            0 Pattern(indices=[1], period=2)
            1 None
            2 Pattern(indices=[1], period=2)
            3 None
            4 Pattern(indices=[1], period=2)
            5 None
            6 Pattern(indices=[1], period=2)
            7 Pattern(indices=[-3, -2, -1])
            8 Pattern(indices=[-3, -2, -1])
            9 Pattern(indices=[-3, -2, -1])

            Matching indices of first pattern offset by ``1``.

            Gets patterns that match next ten indices with rotation set to
            ``1``:

            >>> for i in range(10, 20):
            ...     match = patterns.get_matching_pattern(i, 10, rotation=1)
            ...     print(i, match)
            ...
            10 Pattern(indices=[1], period=2)
            11 None
            12 Pattern(indices=[1], period=2)
            13 None
            14 Pattern(indices=[1], period=2)
            15 None
            16 Pattern(indices=[1], period=2)
            17 None
            18 Pattern(indices=[1], period=2)
            19 None

            Matching indices of first pattern offset by ``1``.

        ..  container:: example

            With inverted patterns:

            >>> patterns = abjad.PatternTuple([
            ...     abjad.Pattern(
            ...         indices=[-3],
            ...         inverted=True,
            ...         ),
            ...     ])

            >>> for i in range(10):
            ...     match = patterns.get_matching_pattern(i, 10)
            ...     print(i, match)
            ...
            0 Pattern(indices=[-3], inverted=True)
            1 Pattern(indices=[-3], inverted=True)
            2 Pattern(indices=[-3], inverted=True)
            3 Pattern(indices=[-3], inverted=True)
            4 Pattern(indices=[-3], inverted=True)
            5 Pattern(indices=[-3], inverted=True)
            6 Pattern(indices=[-3], inverted=True)
            7 None
            8 Pattern(indices=[-3], inverted=True)
            9 Pattern(indices=[-3], inverted=True)

        Returns pattern or none.
        """
        for pattern in reversed(self):
            if hasattr(pattern, 'pattern'):
                if pattern.pattern.matches_index(
                    index, total_length, rotation=rotation):
                    return pattern
            elif pattern.matches_index(index, total_length, rotation=rotation):
                return pattern

    def get_matching_payload(self, index, total_length, rotation=None):
        """
        Gets payload attached to pattern matching ``index``.

        ..  container:: example

            Two patterns. Underlying notes with even divisions
            assigned to the last three indices:

            >>> patterns = abjad.PatternTuple([
            ...     abjad.Pattern(
            ...         indices=[0],
            ...         payload='staccato',
            ...         period=1,
            ...         ),
            ...     abjad.Pattern(
            ...         indices=[-3, -2, -1],
            ...         payload='tenuto',
            ...         ),
            ...     ])

            Over ten indices:

            >>> for i in range(10):
            ...     match = patterns.get_matching_payload(i, 10)
            ...     print(i, match)
            ...
            0 staccato
            1 staccato
            2 staccato
            3 staccato
            4 staccato
            5 staccato
            6 staccato
            7 tenuto
            8 tenuto
            9 tenuto

            Over fifteen indices:

            >>> for i in range(15):
            ...     match = patterns.get_matching_payload(i, 15)
            ...     print(i, match)
            ...
            0 staccato
            1 staccato
            2 staccato
            3 staccato
            4 staccato
            5 staccato
            6 staccato
            7 staccato
            8 staccato
            9 staccato
            10 staccato
            11 staccato
            12 tenuto
            13 tenuto
            14 tenuto

        """
        pattern = self.get_matching_pattern(index, total_length, rotation=rotation)
        payload = None
        if pattern:
            payload = pattern.payload
        return payload

    ### PRIVATE PROPERTIES ###

    @property
    def _item_coercer(self):
        import abjad

        def coerce_(argument):
            if hasattr(argument, 'pattern'):
                pass
            elif not isinstance(argument, abjad.Pattern):
                argument = abjad.Pattern(*argument)
            return argument
        return coerce_



def all_are_nonnegative_integers(argument):
    """
    Is true when ``argument`` is an iterable collection of nonnegative
    integers.

    ..  container:: example

        >>> abjad.mathtools.all_are_nonnegative_integers([0, 1, 2, 99])
        True

        >>> abjad.mathtools.all_are_nonnegative_integers([0, 1, 2, -99])
        False

    Returns true or false.
    """
    from abjad import mathtools
    try:
        return all(mathtools.is_nonnegative_integer(_) for _ in argument)
    except TypeError:
        return False



from .Container import Container


class GraceContainer(Container):
    r"""
    Grace container.

    LilyPond positions grace notes immediately before main notes.

    LilyPond formats grace notes with neither a slashed nor a slur.

    .. container:: example

        Grace notes:

        >>> voice = abjad.Voice("c'4 d'4 e'4 f'4")
        >>> grace_notes = [abjad.Note("c'16"), abjad.Note("d'16")]
        >>> grace_container = abjad.GraceContainer(grace_notes)
        >>> abjad.attach(grace_container, voice[1])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(voice)
            \new Voice
            {
                c'4
                \grace {
                    c'16
                    d'16
                }
                d'4
                e'4
                f'4
            }

    Fill grace containers with notes, rests or chords.

    Attach grace containers to notes, rests or chords.

    ..  container:: example

        Detaches grace container:

        >>> voice = abjad.Voice("c'4 d'4 e'4 f'4")
        >>> note = abjad.Note("cs'16")
        >>> grace_container = abjad.GraceContainer([note])
        >>> abjad.attach(grace_container, voice[1])
        >>> note = abjad.Note("ds'16")
        >>> after_grace_container = abjad.AfterGraceContainer([note])
        >>> abjad.attach(after_grace_container, voice[1])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(voice)
            \new Voice
            {
                c'4
                \grace {
                    cs'16
                }
                \afterGrace
                d'4
                {
                    ds'16
                }
                e'4
                f'4
            }

        >>> abjad.detach(abjad.GraceContainer, voice[1])
        (GraceContainer("cs'16"),)
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(voice)
            \new Voice
            {
                c'4
                \afterGrace
                d'4
                {
                    ds'16
                }
                e'4
                f'4
            }

    ..  container:: example

        Grace containers can be moved:

        >>> staff = abjad.Staff("c'4 d'")
        >>> container = abjad.GraceContainer("fs'32")
        >>> abjad.attach(container, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                \grace {
                    fs'32
                }
                c'4
                d'4
            }

        >>> result = abjad.detach(abjad.GraceContainer, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                c'4
                d'4
            }

        >>> abjad.attach(result[0], staff[1])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                c'4
                \grace {
                    fs'32
                }
                d'4
            }

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Containers'

    __slots__ = (
        '_carrier',
        )

    ### INITIALIZER ###

    def __init__(self, components=None):
        self._carrier = None
        Container.__init__(self, components)

    ### SPECIAL METHODS ###

    def __getnewargs__(self):
        """
        Gets new grace container arguments.

        Returns tuple of single empty list.
        """
        return ([],)

    ### PRIVATE METHODS ###

    def _attach(self, leaf):
        import abjad
        if not isinstance(leaf, abjad.Leaf):
            message = 'must attach to leaf: {!r}.'
            message = message.format(leaf)
            raise TypeError(message)
        leaf._grace_container = self
        self._carrier = leaf

    def _detach(self):
        if self._carrier is not None:
            carrier = self._carrier
            carrier._grace_container = None
            self._carrier = None
        return self

    def _format_open_brackets_slot(self, bundle):
        result = []
        result.append([('grace_brackets', 'open'), [r'\grace {']])
        return tuple(result)

    def _get_lilypond_format(self):
        self._update_now(indicators=True)
        return self._format_component()



from abjad.system.AbjadValueObject import AbjadValueObject
from abjad.pitch.IntervalSegment import IntervalSegment
from abjad.pitch.NamedInterval import NamedInterval
from abjad.system.FormatSpecification import FormatSpecification
from abjad.top.sequence import sequence


class Mode(AbjadValueObject):
    """
    Mode.

    ..  container:: example

        Initializes from string:

        >>> abjad.Mode('major')
        Mode('major')

    ..  container:: example

        Initializes from other mode:

        >>> mode = abjad.Mode('dorian')
        >>> abjad.Mode(mode)
        Mode('dorian')

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_named_interval_segment',
        '_mode_name',
        )

    ### INITIALIZER ###

    def __init__(self, mode_name='dorian'):
        if isinstance(mode_name, str):
            mode_name = mode_name
        elif isinstance(mode_name, Mode):
            mode_name = mode_name.mode_name
        else:
            message = 'must be mode or mode name: {!r}.'
            message = message.format(mode_name)
            raise TypeError(message)
        mdi_segment = self._initialize_with_mode_name(mode_name)
        self._named_interval_segment = mdi_segment
        self._mode_name = mode_name

    ### SPECIAL METHODS ###

    def __eq__(self, argument):
        """
        Is true when ``argument`` is a mode with mode name equal to that of
        this mode.

        ..  container:: example

            >>> mode_1 = abjad.Mode('major')
            >>> mode_2 = abjad.Mode('major')
            >>> mode_3 = abjad.Mode('dorian')

            >>> mode_1 == mode_1
            True
            >>> mode_1 == mode_2
            True
            >>> mode_1 == mode_3
            False

            >>> mode_2 == mode_1
            True
            >>> mode_2 == mode_2
            True
            >>> mode_2 == mode_3
            False

            >>> mode_3 == mode_1
            False
            >>> mode_3 == mode_2
            False
            >>> mode_3 == mode_3
            True

        Returns true or false.
        """
        return super().__eq__(argument)

    def __hash__(self):
        """
        Hashes mode.

        Required to be explicitly redefined on Python 3 if __eq__ changes.

        Returns integer.
        """
        return super().__hash__()

    def __len__(self):
        """
        Length of mode.

        ..  container:: example

            >>> len(abjad.Mode('dorian'))
            7

        Returns nonnegative integer.
        """
        return len(self.named_interval_segment)

    def __str__(self):
        """
        String representation of mode.

        ..  container:: example

            >>> str(abjad.Mode('dorian'))
            'dorian'

        Returns string.
        """
        return self.mode_name

    ### PRIVATE METHODS ###

    def _get_format_specification(self):
        values = [self.mode_name]
        return FormatSpecification(
            client=self,
            repr_is_indented=False,
            storage_format_is_indented=False,
            storage_format_args_values=values,
            )

    def _initialize_with_mode_name(self, mode_name):
        mdi_segment = []
        m2 = NamedInterval('m2')
        M2 = NamedInterval('M2')
        A2 = NamedInterval('aug2')
        dorian = [M2, m2, M2, M2, M2, m2, M2]
        if mode_name == 'dorian':
            mdi_segment.extend(sequence(dorian).rotate(n=0))
        elif mode_name == 'phrygian':
            mdi_segment.extend(sequence(dorian).rotate(n=-1))
        elif mode_name == 'lydian':
            mdi_segment.extend(sequence(dorian).rotate(n=-2))
        elif mode_name == 'mixolydian':
            mdi_segment.extend(sequence(dorian).rotate(n=-3))
        elif mode_name in ('aeolian', 'minor', 'natural minor'):
            mdi_segment.extend(sequence(dorian).rotate(n=-4))
        elif mode_name == 'locrian':
            mdi_segment.extend(sequence(dorian).rotate(n=-5))
        elif mode_name in ('ionian', 'major'):
            mdi_segment.extend(sequence(dorian).rotate(n=-6))
        elif mode_name == 'melodic minor':
            mdi_segment.extend([M2, m2, M2, M2, M2, M2, m2])
        elif mode_name == 'harmonic minor':
            mdi_segment.extend([M2, m2, M2, M2, m2, A2, m2])
        else:
            message = 'unknown mode name: {!r}.'
            message = message.format(mode_name)
            raise ValueError(message)
        return IntervalSegment(
            items=mdi_segment,
            item_class=NamedInterval,
            )

    ### PUBLIC PROPERTIES ###

    @property
    def mode_name(self):
        """
        Gets mode name.

        ..  container:: example

            >>> abjad.Mode('major').mode_name
            'major'

            >>> abjad.Mode('dorian').mode_name
            'dorian'

        Returns string.
        """
        return self._mode_name

    @property
    def named_interval_segment(self):
        """
        Gets named interval segment.

        ..  container:: example

            >>> mode = abjad.Mode('major')
            >>> str(mode.named_interval_segment)
            '<+M2, +M2, +m2, +M2, +M2, +M2, +m2>'

            >>> mode = abjad.Mode('dorian')
            >>> str(mode.named_interval_segment)
            '<+M2, +m2, +M2, +M2, +M2, +m2, +M2>'

        Returns named interval segment.
        """
        return self._named_interval_segment



def show(argument, return_timing=False, **keywords):
    r"""
    Shows ``argument``.

    ..  container:: example

        Shows note:

        >>> note = abjad.Note("c'4")
        >>> abjad.show(note) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(note)
            c'4

    ..  container:: example

        Shows staff:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                c'4
                d'4
                e'4
                f'4
            }

    Makes LilyPond input files and output PDF.

    Writes LilyPond input file and output PDF to Abjad output directory.

    Opens output PDF.

    Returns none when ``return_timing`` is false.

    Returns pair of ``abjad_formatting_time`` and ``lilypond_rendering_time``
    when ``return_timing`` is true.
    """
    import abjad
    if not hasattr(argument, '__illustrate__'):
        message = 'must have __illustrate__ method: {!r}.'
        message = message.format(argument)
        raise Exception(message)
    result = abjad.persist(argument).as_pdf(**keywords)
    pdf_file_path = result[0]
    abjad_formatting_time = result[1]
    lilypond_rendering_time = result[2]
    success = result[3]
    if success:
        abjad.IOManager.open_file(pdf_file_path)
    else:
        with open(abjad.abjad_configuration.lilypond_log_file_path, 'r') as fp:
            print(fp.read())
    if return_timing:
        return abjad_formatting_time, lilypond_rendering_time



"""
Utility classes and functions.
"""

from abjad import mathtools  # noqa
Infinity = mathtools.Infinity()
NegativeInfinity = mathtools.NegativeInfinity()
del mathtools

from .CyclicTuple import CyclicTuple
from .Duration import Duration
from .Enumerator import Enumerator
from .Inequality import Inequality
from .DurationInequality import DurationInequality
from .Expression import Expression
from .LengthInequality import LengthInequality
from .Multiplier import Multiplier
from .Offset import Offset
from .Pattern import Pattern
from .TypedCollection import TypedCollection
from .TypedTuple import TypedTuple
from .OrderedDict import OrderedDict
from .PatternTuple import PatternTuple
from .PitchInequality import PitchInequality
from .Sequence import Sequence
from .SortedCollection import SortedCollection
from .String import String
from .TypedCounter import TypedCounter
from .TypedFrozenset import TypedFrozenset
from .TypedList import TypedList

from .compare_images import compare_images
from .list_all_classes import list_all_classes
from .list_all_functions import list_all_functions
from .list_all_ide_classes import list_all_ide_classes
from .list_all_ide_functions import list_all_ide_functions
from .yield_all_modules import yield_all_modules



import typing
from abjad.system.AbjadValueObject import AbjadValueObject
from abjad.pitch.NamedPitch import NamedPitch
from abjad.pitch.StaffPosition import StaffPosition
from abjad.system.FormatSpecification import FormatSpecification
from abjad.system.LilyPondFormatBundle import LilyPondFormatBundle
from abjad.top.iterate import iterate


class Clef(AbjadValueObject):
    r"""
    Clef.

    ..  container:: example

        Some available clefs:

        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8 g'8 a'8 b'8 c''8")
        >>> clef = abjad.Clef('treble')
        >>> abjad.attach(clef, staff[0])
        >>> clef = abjad.Clef('alto')
        >>> abjad.attach(clef, staff[1])
        >>> clef = abjad.Clef('bass')
        >>> abjad.attach(clef, staff[2])
        >>> clef = abjad.Clef('treble^8')
        >>> abjad.attach(clef, staff[3])
        >>> clef = abjad.Clef('bass_8')
        >>> abjad.attach(clef, staff[4])
        >>> clef = abjad.Clef('tenor')
        >>> abjad.attach(clef, staff[5])
        >>> clef = abjad.Clef('bass^15')
        >>> abjad.attach(clef, staff[6])
        >>> clef = abjad.Clef('percussion')
        >>> abjad.attach(clef, staff[7])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                \clef "treble"
                c'8
                \clef "alto"
                d'8
                \clef "bass"
                e'8
                \clef "treble^8"
                f'8
                \clef "bass_8"
                g'8
                \clef "tenor"
                a'8
                \clef "bass^15"
                b'8
                \clef "percussion"
                c''8
            }

    ..  container:: example

        Clefs can be tagged:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> abjad.attach(abjad.Clef('treble'), staff[0], tag='RED')
        >>> abjad.show(staff) # doctest: +SKIP

        >>> abjad.f(staff)
        \new Staff
        {
            \clef "treble" %! RED
            c'4
            d'4
            e'4
            f'4
        }

    ..  container:: example

        LilyPond can not handle simultaneous clefs:

        >>> voice_1 = abjad.Voice("e'8 g' f' a' g' b'")
        >>> abjad.attach(abjad.Clef('treble'), voice_1[0], context='Voice')
        >>> abjad.attach(abjad.LilyPondLiteral(r'\voiceOne'), voice_1)
        >>> voice_1.consists_commands.append('Clef_engraver')
        >>> voice_2 = abjad.Voice("c'4. c,8 b,, a,,")
        >>> abjad.attach(abjad.Clef('treble'), voice_2[0], context='Voice')
        >>> abjad.attach(abjad.Clef('bass'), voice_2[1], context='Voice')
        >>> abjad.attach(abjad.LilyPondLiteral(r'\voiceTwo'), voice_2)
        >>> voice_2.consists_commands.append('Clef_engraver')
        >>> staff = abjad.Staff([voice_1, voice_2], is_simultaneous=True)
        >>> staff.remove_commands.append('Clef_engraver')
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            \with
            {
                \remove Clef_engraver
            }
            <<
                \new Voice
                \with
                {
                    \consists Clef_engraver
                }
                {
                    \voiceOne
                    \clef "treble"
                    e'8
                    g'8
                    f'8
                    a'8
                    g'8
                    b'8
                }
                \new Voice
                \with
                {
                    \consists Clef_engraver
                }
                {
                    \voiceTwo
                    \clef "treble"
                    c'4.
                    \clef "bass"
                    c,8
                    b,,8
                    a,,8
                }
            >>

        But Abjad components work fine:

        >>> for leaf in abjad.select(voice_1).leaves():
        ...     leaf, abjad.inspect(leaf).effective(abjad.Clef)
        ...
        (Note("e'8"), Clef('treble'))
        (Note("g'8"), Clef('treble'))
        (Note("f'8"), Clef('treble'))
        (Note("a'8"), Clef('treble'))
        (Note("g'8"), Clef('treble'))
        (Note("b'8"), Clef('treble'))

        >>> for leaf in abjad.select(voice_2).leaves():
        ...     leaf, abjad.inspect(leaf).effective(abjad.Clef)
        ...
        (Note("c'4."), Clef('treble'))
        (Note('c,8'), Clef('bass'))
        (Note('b,,8'), Clef('bass'))
        (Note('a,,8'), Clef('bass'))

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_hide',
        '_middle_c_position',
        '_name',
        )

    _clef_name_to_middle_c_position = {
        'treble': -6,
        'alto': 0,
        'tenor': 2,
        'bass': 6,
        'french': -8,
        'soprano': -4,
        'mezzosoprano': -2,
        'baritone': 4,
        'varbaritone': 4,
        'percussion': 0,
        'tab': 0,
        }


    _context = 'Staff'

    _format_slot = 'opening'

    _parameter = True

    _redraw = True

    _to_width = {
        'alto': 2.75,
        'bass': 2.75,
        'percussion': 2.5,
        'tenor': 2.75,
        'treble': 2.5,
        }

    ### INITIALIZER ###

    def __init__(
        self,
        name: str = 'treble',
        *,
        hide: bool = None,
        ) -> None:
        if isinstance(name, str):
            self._name = name
        elif isinstance(name, type(self)):
            self._name = name.name
        else:
            raise TypeError('can not initialize clef: {name!r}.')
        if hide is not None:
            hide = bool(hide)
        self._hide = hide
        middle_c_position = self._calculate_middle_c_position(self._name)
        middle_c_position = StaffPosition(middle_c_position)
        self._middle_c_position = middle_c_position

    ### SPECIAL METHODS ###

    def __format__(self, format_specification='') -> str:
        r"""
        Formats clef.

        ..  container:: example

            >>> clef = abjad.Clef('treble')
            >>> print(format(clef))
            abjad.Clef('treble')

        ..  container:: example

            >>> clef = abjad.Clef('treble')
            >>> print(format(clef, 'lilypond'))
            \clef "treble"

        Returns string.
        """
        if format_specification == 'lilypond':
            return self._get_lilypond_format()
        return super().__format__(
            format_specification=format_specification,
            )

    ### PRIVATE METHODS ###

    def _calculate_middle_c_position(self, clef_name):
        alteration = 0
        if '_' in self._name:
            base_name, part, suffix = clef_name.partition('_')
            if suffix == '8':
                alteration = 7
            elif suffix == '15':
                alteration = 13
            else:
                message = 'bad clef alteration suffix: {!r}.'
                message = message.format(suffix)
                raise Exception(message)
        elif '^' in self._name:
            base_name, part, suffix = clef_name.partition('^')
            if suffix == '8':
                alteration = -7
            elif suffix == '15':
                alteration = -13
            else:
                message = "bad clef alteration suffix: {!r}."
                message = message.format(suffix)
                raise Exception(message)
        else:
            base_name = clef_name
        return self._clef_name_to_middle_c_position[base_name] + alteration

    def _clef_name_to_staff_position_zero(self, clef_name):
        return {
            'treble': NamedPitch('B4'),
            'alto': NamedPitch('C4'),
            'tenor': NamedPitch('A3'),
            'bass': NamedPitch('D3'),
            'french': NamedPitch('D5'),
            'soprano': NamedPitch('G4'),
            'mezzosoprano': NamedPitch('E4'),
            'baritone': NamedPitch('F3'),
            'varbaritone': NamedPitch('F3'),
            'percussion': None,
            'tab': None,
            }[clef_name]

    def _get_format_specification(self):
        return FormatSpecification(
            self,
            repr_is_indented=False,
            storage_format_args_values=[self.name],
            storage_format_is_indented=False,
            )

    def _get_lilypond_format(self):
        return rf'\clef "{self.name}"'

    def _get_lilypond_format_bundle(self, component=None):
        bundle = LilyPondFormatBundle()
        if not self.hide:
            bundle.opening.commands.append(self._get_lilypond_format())
        return bundle

    @classmethod
    def _list_clef_names(class_):
        return list(sorted(class_._clef_name_to_middle_c_position))

    ### PUBLIC METHODS ###

    @staticmethod
    def from_selection(selection) -> 'Clef':
        """
        Makes clef from ``selection``.

        ..  container:: example

            >>> maker = abjad.NoteMaker()
            >>> notes = maker(range(-12, -6), [(1, 4)])
            >>> staff = abjad.Staff(notes)
            >>> abjad.Clef.from_selection(staff)
            Clef('bass')

            Choses between treble and bass based on minimal number of ledger
            lines.

        """
        pitches = iterate(selection).pitches()
        diatonic_pitch_numbers = [
            pitch._get_diatonic_pitch_number() for pitch in pitches
            ]
        max_diatonic_pitch_number = max(diatonic_pitch_numbers)
        min_diatonic_pitch_number = min(diatonic_pitch_numbers)
        lowest_treble_line_pitch = NamedPitch('E4')
        lowest_treble_line_diatonic_pitch_number = \
            lowest_treble_line_pitch._get_diatonic_pitch_number()
        candidate_steps_below_treble = \
            lowest_treble_line_diatonic_pitch_number - \
            min_diatonic_pitch_number
        highest_bass_line_pitch = NamedPitch('A3')
        highest_bass_line_diatonic_pitch_number = \
            highest_bass_line_pitch._get_diatonic_pitch_number()
        candidate_steps_above_bass = \
            max_diatonic_pitch_number - highest_bass_line_diatonic_pitch_number
        if candidate_steps_above_bass < candidate_steps_below_treble:
            return Clef('bass')
        else:
            return Clef('treble')

    ### PUBLIC PROPERTIES ###

    @property
    def context(self) -> str:
        """
        Gets (historically conventional) context.

        ..  container:: example

            >>> clef = abjad.Clef('treble')
            >>> clef.context
            'Staff'

        Override with ``abjad.attach(..., context='...')``.
        """
        return self._context

    @property
    def hide(self) -> typing.Optional[bool]:
        r"""
        Is true when clef should not appear in output (but should still
        determine effective clef).

        ..  container:: example

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> abjad.attach(abjad.Clef('treble'), staff[0]) 
            >>> abjad.attach(abjad.Clef('alto', hide=True), staff[2]) 
            >>> abjad.show(staff) # doctest: +SKIP

            >>> abjad.f(staff)
            \new Staff
            {
                \clef "treble"
                c'4
                d'4
                e'4
                f'4
            }

            >>> for leaf in abjad.iterate(staff).leaves():
            ...     leaf, abjad.inspect(leaf).effective(abjad.Clef)
            ...
            (Note("c'4"), Clef('treble'))
            (Note("d'4"), Clef('treble'))
            (Note("e'4"), Clef('alto', hide=True))
            (Note("f'4"), Clef('alto', hide=True))

        """
        return self._hide

    @property
    def middle_c_position(self) -> int:
        """
        Gets middle C position of clef.

        ..  container:: example

            Gets staff position of middle C in treble clef:

            >>> abjad.Clef('treble').middle_c_position
            StaffPosition(-6)

        ..  container:: example

            Gets staff position of middle C in alto clef:

            >>> abjad.Clef('alto').middle_c_position
            StaffPosition(0)

        """
        return self._middle_c_position

    @property
    def name(self) -> str:
        """
        Gets name of clef.

        ..  container:: example

            Gets name treble clef:

            >>> abjad.Clef('treble').name
            'treble'

        ..  container:: example

            Gets name of alto clef:

            >>> abjad.Clef('alto').name
            'alto'

        """
        return self._name

    @property
    def parameter(self) -> bool:
        """
        Is true.

        ..  container:: example

            >>> abjad.Clef('treble').parameter
            True

        Class constant.
        """
        return self._parameter
        
    @property
    def redraw(self) -> bool:
        """
        Is true.

        ..  container:: example

            >>> abjad.Clef('treble').redraw
            True

        Class constant.
        """
        return self._redraw

    @property
    def tweaks(self) -> None:
        r"""
        Are not implemented on clef.
        
        The LilyPond ``\clef`` command refuses tweaks.

        Override the LilyPond ``Clef`` grob instead.
        """
        pass



def persist(client):
    r"""
    Makes persistence manager.

    ..  container:: example

        Persists staff as LilyPond file:

        >>> staff = abjad.Staff("c'4 e'4 d'4 f'4")
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                c'4
                e'4
                d'4
                f'4
            }

        >>> abjad.persist(staff).as_ly() # doctest: +SKIP

    ..  container:: example

        Returns persistence agent:

        >>> abjad.persist(staff)
        PersistenceManager(client=Staff("c'4 e'4 d'4 f'4"))

    """
    import abjad
    return abjad.PersistenceManager(client)



from abjad import core
from .Music import Music


class SequentialMusic(Music):
    """
    Abjad model of the LilyPond AST sequential music node.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### PUBLIC METHODS ###

    def construct(self):
        """
        Constructs sequential music.

        Returns Abjad container.
        """
        container = core.Container()
        for x in self.music:
            if isinstance(x, core.Component):
                container.append(x)
            elif isinstance(x, type(self)):
                container.extend(x.construct())
        return container



import abc
import functools
import numbers
from abjad import mathtools
from abjad.system.AbjadValueObject import AbjadValueObject
from . import constants


@functools.total_ordering
class PitchClass(AbjadValueObject):
    """
    Abstract pitch-class.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    @abc.abstractmethod
    def __init__(self, argument):
        import abjad
        if isinstance(argument, str):
            match = constants._comprehensive_pitch_name_regex.match(argument)
            if not match:
                match = constants._comprehensive_pitch_class_name_regex.match(argument)
            if not match:
                message = 'can not instantiate {} from {!r}.'
                message = message.format(type(self).__name__, argument)
                raise ValueError(message)
            group_dict = match.groupdict()
            dpc_name = group_dict['diatonic_pc_name'].lower()
            dpc_number = constants._diatonic_pc_name_to_diatonic_pc_number[dpc_name]
            alteration = abjad.Accidental(group_dict['comprehensive_accidental']).semitones
            self._from_named_parts(dpc_number, alteration)
        elif isinstance(argument, numbers.Number):
            self._from_number(argument)
        elif isinstance(argument, (abjad.Pitch, abjad.PitchClass)):
            self._from_pitch_or_pitch_class(argument)
        else:
            try:
                pitch = abjad.NamedPitch(argument)
                self._from_pitch_or_pitch_class(pitch)
            except Exception:
                message = 'can not instantiate {} from {!r}.'
                message = message.format(type(self).__name__, argument)
                raise ValueError(message)

    ### SPECIAL METHODS ###

    def __float__(self):
        """
        Coerce to float.

        Returns float.
        """
        return float(self.number)

    def __format__(self, format_specification=''):
        """
        Formats pitch-class.

        Set `format_specification` to `''`, `'lilypond'` or `'storage'`.

        Returns string.
        """
        if format_specification == 'lilypond':
            return self._get_lilypond_format()
        return super().__format__(format_specification=format_specification)

    @abc.abstractmethod
    def __lt__(self, argument):
        """
        Is true when pitch-class is less than `argument`.

        Returns true or false.
        """
        raise NotImplementedError

    ### PRIVATE METHODS ###

    @abc.abstractmethod
    def _from_named_parts(self, dpc_number, alteration):
        raise NotImplementedError

    @abc.abstractmethod
    def _from_number(self, number):
        raise NotImplementedError

    @abc.abstractmethod
    def _from_pitch_or_pitch_class(self, pitch_or_pitch_class):
        raise NotImplementedError

    @abc.abstractmethod
    def _get_alteration(self):
        raise NotImplementedError

    @abc.abstractmethod
    def _get_diatonic_pc_number(self):
        raise NotImplementedError

    @abc.abstractmethod
    def _get_lilypond_format(self):
        raise NotImplementedError

    @staticmethod
    def _to_nearest_quarter_tone(number):
        number = round((float(number) % 12) * 4) / 4
        div, mod = divmod(number, 1)
        if mod == 0.75:
            div += 1
        elif mod == 0.5:
            div += 0.5
        div %= 12
        return mathtools.integer_equivalent_number_to_integer(div)

    ### PUBLIC PROPERTIES ###

    @abc.abstractproperty
    def accidental(self):
        """
        Gets accidental of pitch-class.
        """
        raise NotImplementedError

    @abc.abstractproperty
    def pitch_class_label(self):
        """
        Gets pitch-class label of pitch-class.
        """
        raise NotImplementedError

    ### PUBLIC METHODS ###

    @abc.abstractmethod
    def invert(self, axis=None):
        """
        Inverts pitch-class about `axis`.

        Returns new pitch-class.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def multiply(self, n=1):
        """
        Multiplies pitch-class by `n`.

        Returns new pitch-class.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def transpose(self, n=0):
        """
        Transposes pitch-class by index `n`.

        Returns new pitch-class.
        """
        raise NotImplementedError



import typing

_lilypond_parsers_by_language: typing.Dict = {}


def parse(string, language='english'):
    r"""
    Parses LilyPond ``string``.

    ..  container:: example

        Parses LilyPond string with English note names:

        >>> container = abjad.parse("{c'4 d'4 e'4 f'4}")
        >>> abjad.show(container) # doctest: +SKIP

    ..  container:: example

        Parses LilyPond string with Dutch note names:

        >>> container = abjad.parse(
        ...     "{c'8 des' e' fis'}",
        ...     language='nederlands',
        ...     )
        >>> abjad.show(container) # doctest: +SKIP

    Returns Abjad component.
    """
    import abjad.parser
    import abjad.rhythmtrees
    if string.startswith('abj:'):
        return abjad.parser.parse_reduced_ly_syntax(string[4:])
    elif string.startswith('rtm:'):
        return abjad.rhythmtrees.parse_rtm_syntax(string[4:])
    if language not in _lilypond_parsers_by_language:
        parser = abjad.parser.LilyPondParser(default_language=language)
        _lilypond_parsers_by_language[language] = parser
    return _lilypond_parsers_by_language[language](string)



def attach(
    attachable,
    target,
    context=None,
    deactivate=None,
    synthetic_offset=None,
    tag=None,
    wrapper=None,
    ):
    r"""
    Attaches ``attachable`` to ``target``.
    
    First form attaches indicator ``attachable`` to single leaf ``target``.
    
    Second form attaches spanner ``attachable`` to leaf selection ``target``.

    Third for attaches grace container ``attachable`` to leaf ``target``.

    Fourth form attaches time signature ``attachable`` to measure ``target``.

    Fifth form attaches wrapper ``attachable`` to unknown (?) ``target``.

    ..  container:: example

        Attaches clef to first note in staff:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> abjad.attach(abjad.Clef('alto'), staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                \clef "alto"
                c'4
                d'4
                e'4
                f'4
            }

    ..  container:: example

        Attaches accent to last note in staff:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> abjad.attach(abjad.Articulation('>'), staff[-1])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                c'4
                d'4
                e'4
                f'4
                -\accent
            }

    ..  container:: example

        Works with context names:

        >>> voice = abjad.Voice("c'4 d' e' f'", name='MusicVoice')
        >>> staff = abjad.Staff([voice], name='MusicStaff')
        >>> abjad.attach(abjad.Clef('alto'), voice[0], context='MusicStaff')
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \context Staff = "MusicStaff"
            {
                \context Voice = "MusicVoice"
                {
                    \clef "alto"
                    c'4
                    d'4
                    e'4
                    f'4
                }
            }

        >>> for leaf in abjad.select(staff).leaves():
        ...     leaf, abjad.inspect(leaf).effective(abjad.Clef)
        ...
        (Note("c'4"), Clef('alto'))
        (Note("d'4"), Clef('alto'))
        (Note("e'4"), Clef('alto'))
        (Note("f'4"), Clef('alto'))

        Derives context from default ``attachable`` context when ``context`` is
        none.

    ..  container:: example

        Two contexted indicators can not be attached at the same offset if both
        indicators are active:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> abjad.attach(abjad.Clef('treble'), staff[0])
        >>> abjad.attach(abjad.Clef('alto'), staff[0])
        Traceback (most recent call last):
            ...
        abjad...PersistentIndicatorError: Can not attach ...

        But simultaneous contexted indicators are allowed if only one is active
        (and all others are inactive):

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> abjad.attach(abjad.Clef('treble'), staff[0])
        >>> abjad.attach(
        ...     abjad.Clef('alto'),
        ...     staff[0],
        ...     deactivate=True,
        ...     tag='+PARTS_1',
        ...     )
        >>> abjad.attach(
        ...     abjad.Clef('tenor'),
        ...     staff[0],
        ...     deactivate=True,
        ...     tag='+PARTS_2',
        ...     )
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                \clef "treble"
            %@% \clef "alto" %! +PARTS_1
            %@% \clef "tenor" %! +PARTS_2
                c'4
                d'4
                e'4
                f'4
            }

        Active indicator is always effective when competing inactive indicators
        are present:

        >>> for note in staff:
        ...     clef = abjad.inspect(staff[0]).effective(abjad.Clef)
        ...     note, clef
        ...
        (Note("c'4"), Clef('treble'))
        (Note("d'4"), Clef('treble'))
        (Note("e'4"), Clef('treble'))
        (Note("f'4"), Clef('treble'))

        But a lone inactivate indicator is effective when no active indicator
        is present:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> abjad.attach(
        ...     abjad.Clef('alto'),
        ...     staff[0],
        ...     deactivate=True,
        ...     tag='+PARTS',
        ...     )
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
            %@% \clef "alto" %! +PARTS
                c'4
                d'4
                e'4
                f'4
            }

        >>> for note in staff:
        ...     clef = abjad.inspect(staff[0]).effective(abjad.Clef)
        ...     note, clef
        ...
        (Note("c'4"), Clef('alto'))
        (Note("d'4"), Clef('alto'))
        (Note("e'4"), Clef('alto'))
        (Note("f'4"), Clef('alto'))

    ..  container:: example

        Tag must exist when ``deactivate`` is true:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> abjad.attach(abjad.Clef('alto'), staff[0], deactivate=True)
        Traceback (most recent call last):
            ...
        Exception: tag must exist when deactivate is true.

    ..  container:: example

        Returns wrapper when ``wrapper`` is true:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> wrapper = abjad.attach(abjad.Clef('alto'), staff[0], wrapper=True)
        >>> abjad.f(wrapper)
        abjad.Wrapper(
            context='Staff',
            indicator=abjad.Clef('alto'),
            tag=abjad.Tag(),
            )

    Otherwise returns none.
    """
    import abjad

    assert attachable is not None, repr(attachable)
    assert target is not None, repr(target)

    nonindicator_prototype = (
        abjad.AfterGraceContainer,
        abjad.GraceContainer,
        abjad.Spanner,
        )

    if context is not None and isinstance(attachable, nonindicator_prototype):
        message = 'set context only for indicators, not {!r}.'
        message = message.format(attachable)
        raise Exception(message)

    if deactivate is True and tag is None:
        raise Exception(f'tag must exist when deactivate is true.')

    if hasattr(attachable, '_before_attach'):
        attachable._before_attach(target)

    if hasattr(attachable, '_attachment_test_all'):
        result = attachable._attachment_test_all(target)
        if result is not True:
            assert isinstance(result, list), repr(result)
            result = ['  ' + _ for _ in result]
            message = f'{attachable!r}._attachment_test_all():'
            result.insert(0, message)
            message = '\n'.join(result)
            raise Exception(message)

    grace_container = (abjad.AfterGraceContainer, abjad.GraceContainer)
    if isinstance(attachable, abjad.Spanner):
        if not isinstance(target, abjad.Selection):
            raise Exception('spanners attach to leaf selections only.')
        if not target.are_leaves():
            raise Exception('spanners attach to leaf selections only.')
        attachable._attach(
            target,
            deactivate=deactivate,
            tag=tag,
            )
        return
    elif isinstance(attachable, grace_container):
        if not isinstance(target, abjad.Leaf):
            raise Exception('grace containers attach to single leaf only.')
        attachable._attach(target)
        return

    assert isinstance(target, abjad.Component), repr(target)

    if isinstance(target, abjad.Container):
        acceptable = False
        if isinstance(attachable, (dict, str, abjad.Wrapper)):
            acceptable = True
        if (isinstance(attachable, abjad.TimeSignature) and
            isinstance(target, abjad.Measure)):
            acceptable = True
        if getattr(attachable, '_can_attach_to_containers', False):
            acceptable = True
        if not acceptable:
            message = 'can not attach {!r} to containers: {!r}'
            message = message.format(attachable, target)
            raise Exception(message)
    elif not isinstance(target, abjad.Leaf):
        message = 'indicator {!r} must attach to leaf instead, not {!r}.'
        message = message.format(attachable, target)
        raise Exception(message)

    component = target
    assert isinstance(component, abjad.Component)

    annotation = None
    if isinstance(attachable, abjad.Wrapper):
        annotation = attachable.annotation
        context = context or attachable.context
        deactivate = deactivate or attachable.deactivate
        synthetic_offset = synthetic_offset or attachable.synthetic_offset
        tag = tag or attachable.tag
        attachable._detach()
        attachable = attachable.indicator

    if hasattr(attachable, 'context'):
        context = context or attachable.context

    wrapper_ = abjad.Wrapper(
        annotation=annotation,
        component=component,
        context=context,
        deactivate=deactivate,
        indicator=attachable,
        synthetic_offset=synthetic_offset,
        tag=tag,
        )
    wrapper_._bind_to_component(component)

    if wrapper is True:
        return wrapper_



from abjad import enums


def tweak(argument):
    r"""
    Makes LilyPond tweak manager.

    ..  container:: example

        Tweaks markup:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> markup = abjad.Markup('Allegro assai', direction=abjad.Up)
        >>> abjad.tweak(markup).color = 'red'
        >>> abjad.attach(markup, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                c'4
                - \tweak color #red
                ^ \markup { "Allegro assai" }
                d'4
                e'4
                f'4
            }

        Survives copy:

        >>> import copy
        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> markup_1 = abjad.Markup('Allegro assai', direction=abjad.Up)
        >>> abjad.tweak(markup_1).color = 'red'
        >>> markup_2 = copy.copy(markup_1)
        >>> abjad.attach(markup_2, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                c'4
                - \tweak color #red
                ^ \markup { "Allegro assai" }
                d'4
                e'4
                f'4
            }

        Survives dot-chaining:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> markup = abjad.Markup('Allegro assai', direction=abjad.Up)
        >>> abjad.tweak(markup).color = 'red'
        >>> markup = markup.italic()
        >>> abjad.attach(markup, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                c'4
                - \tweak color #red
                ^ \markup {
                    \italic
                        "Allegro assai"
                    }
                d'4
                e'4
                f'4
            }

        Works for opposite-directed coincident markup:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> markup_1 = abjad.Markup('Allegro assai ...', direction=abjad.Up)
        >>> abjad.tweak(markup_1).color = 'red'
        >>> abjad.attach(markup_1, staff[0])
        >>> markup_2 = abjad.Markup('... ma non troppo', direction=abjad.Down)
        >>> abjad.tweak(markup_2).color = 'blue'
        >>> abjad.tweak(markup_2).staff_padding = 4
        >>> abjad.attach(markup_2, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                c'4
                - \tweak color #red
                ^ \markup { "Allegro assai ..." }
                - \tweak color #blue
                - \tweak staff-padding #4
                _ \markup { "... ma non troppo" }
                d'4
                e'4
                f'4
            }

        Ignored for same-directed coincident markup:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> markup_1 = abjad.Markup('Allegro assai ...', direction=abjad.Up)
        >>> abjad.tweak(markup_1).color = 'red'
        >>> abjad.attach(markup_1, staff[0])
        >>> markup_2 = abjad.Markup('... ma non troppo', direction=abjad.Up)
        >>> abjad.tweak(markup_2).color = 'blue'
        >>> abjad.tweak(markup_2).staff_padding = 4
        >>> abjad.attach(markup_2, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                c'4
                - \tweak color #red
                ^ \markup { "Allegro assai ..." }
                - \tweak color #blue
                - \tweak staff-padding #4
                ^ \markup { "... ma non troppo" }
                d'4
                e'4
                f'4
            }

    ..  container:: example

        Tweaks note-head:

        >>> staff = abjad.Staff("c'4 cs' d' ds'")
        >>> abjad.tweak(staff[1].note_head).color = 'red'
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                c'4
                \tweak color #red
                cs'4
                d'4
                ds'4
            }

        Tweaks grob aggregated to note-head:

        >>> staff = abjad.Staff("c'4 cs' d' ds'")
        >>> abjad.tweak(staff[1].note_head).accidental.color = 'red'
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                c'4
                \tweak Accidental.color #red
                cs'4
                d'4
                ds'4
            }

    ..  container:: example

        Returns LilyPond tweak manager:

        >>> abjad.tweak(markup_1)
        LilyPondTweakManager(('color', 'red'))

    ..  container:: example

        Tweak expressions work like this:

        >>> abjad.tweak('red').color
        LilyPondTweakManager(('color', 'red'))

        >>> abjad.tweak(6).Y_offset
        LilyPondTweakManager(('Y_offset', 6))

        >>> abjad.tweak(False).bound_details__left_broken__text
        LilyPondTweakManager(('bound_details__left_broken__text', False))

    """
    import abjad
    constants = (enums.Down, enums.Left, enums.Right, enums.Up)
    prototype = (bool, int, float, str, tuple, abjad.Scheme)
    if argument in constants or isinstance(argument, prototype):
        manager = abjad.LilyPondTweakManager()
        manager._pending_value = argument
        return manager
    if not hasattr(argument, '_tweaks'):
        name = type(argument).__name__
        raise NotImplementedError(f'{name} does not allow tweaks (yet).')
    if argument._tweaks is None:
        argument._tweaks = abjad.LilyPondTweakManager()
    return argument._tweaks
